"""Tests for the creation and comparison of Model Metrics"""

import pytest
from pprint import pprint

# Workbench Imports
from workbench.api.feature_set import FeatureSet
from workbench.api.model import Model, ModelType
from workbench.api.endpoint import Endpoint


model_reg = Model("abalone-regression")
model_class = Model("wine-classification")


# Test the Model Metrics
def test_list_inference_runs():
    """Test the List Inference Runs"""
    print("\n\n*** List Inference Runs ***")
    pprint(model_reg.list_inference_runs())
    pprint(model_class.list_inference_runs())


def test_performance_metrics():
    """Test the Model Performance Metrics"""
    print("\n\n*** Performance Metrics ***")
    pprint(model_reg.get_inference_metadata())
    pprint(model_reg.get_inference_metrics())
    pprint(model_class.get_inference_metrics())


def test_retrieval_with_capture_name():
    """Test the retrieval of the model metrics using capture Name"""
    capture_list = model_class.list_inference_runs()
    for capture_name in capture_list:
        print(f"\n\n*** Retrieval with Capture Name ({capture_name}) ***")
        pprint(model_class.get_inference_metadata(capture_name).head())
        pprint(model_class.get_inference_metrics(capture_name).head())
        pprint(model_class.get_inference_predictions(capture_name).head())
        pprint(model_class.confusion_matrix(capture_name))


def test_validation_predictions():
    print("\n\n*** Validation Predictions ***")
    reg_val_preds = model_reg._get_validation_predictions()
    pprint(reg_val_preds.head())
    class_val_preds = model_class._get_validation_predictions()
    pprint(class_val_preds.head())


def test_inference_predictions():
    print("\n\n*** Inference Predictions ***")

    # Make sure we have inference predictions
    end = Endpoint("abalone-regression")
    end.auto_inference(capture=True)
    end = Endpoint("wine-classification")
    end.auto_inference(capture=True)

    # Retrieve the inference predictions
    model_reg = Model("abalone-regression")
    if model_reg.get_inference_predictions() is None:
        print(f"Model {model_reg.name} has no inference predictions!")
        exit(1)
    pprint(model_reg.get_inference_predictions().head())
    if model_class.get_inference_predictions() is None:
        print(f"Model {model_class.name} has no inference predictions!")
        exit(1)
    pprint(model_class.get_inference_predictions().head())


def test_confusion_matrix():
    print("\n\n*** Confusion Matrix ***")
    pprint(model_reg.confusion_matrix())
    pprint(model_class.confusion_matrix())


def test_shap_values():
    print("\n\n*** SHAP Features (regression) ***")
    shap_features = model_reg.shap_importance()
    if shap_features is None:
        print(f"Model {model_reg.name} has no SHAP features!")
    else:
        pprint(shap_features)

    print("\n\n*** SHAP Data (regression) ***")
    shap_data = model_reg.shap_data()
    if shap_data is None:
        print(f"Model {model_reg.name} has no SHAP data!")
    else:
        pprint(shap_features)

    print("\n\n*** SHAP Features (classification) ***")
    shap_features = model_class.shap_importance()
    if shap_features is None:
        print(f"Model {model_class.name} has no SHAP features!")
    else:
        pprint(shap_features)

    print("\n\n*** SHAP Data (classification) ***")
    shap_data = model_class.shap_data()
    if shap_data is None:
        print(f"Model {model_class.name} has no SHAP data!")
    else:
        # Classifiers have a dictionary of dataframes
        for key, df in shap_data.items():
            print(f"{key}")
            print(df.head())


def test_metrics_with_capture_name():
    """Test the Performance Metrics using a Capture Name"""
    metrics = model_reg.get_inference_metrics("auto_inference")
    print("\n\n*** Performance Metrics with Capture Name ***")
    pprint(metrics)
    metrics = model_class.get_inference_metrics("auto_inference")
    pprint(metrics)


@pytest.mark.long
def test_auto_inference():
    # Run auto_inference (back track to FeatureSet)
    my_endpoint = Endpoint("abalone-regression")
    pred_results = my_endpoint.auto_inference()
    print("\n\n*** Auto Inference ***")
    pprint(pred_results.head())

    my_endpoint = Endpoint("wine-classification")
    pred_results = my_endpoint.auto_inference()
    pprint(pred_results.head())


@pytest.mark.long
def test_inference_with_capture_name():
    # Run inference on the model
    capture_name = "my_holdout_test"

    # Grab a dataframe for inference
    my_features = FeatureSet("abalone_features")
    table = my_features.view("training").table
    df = my_features.query(f'SELECT * FROM "{table}" where training = FALSE')

    # Run inference
    my_endpoint = Endpoint("abalone-regression")
    pred_results = my_endpoint.inference(df, capture_name)
    pprint(pred_results.head())


def create_model_and_endpoint():
    # Retrieve an existing FeatureSet
    my_features = FeatureSet("abalone_features")

    # Create a Model/Endpoint from the FeatureSet
    model_reg = my_features.to_model(
        name="abalone-regression", model_type=ModelType.REGRESSOR, target_column="class_number_of_rings"
    )
    model_reg.to_endpoint(name="abalone-regression", tags=["abalone", "public"])


if __name__ == "__main__":

    # Set Pandas display options
    import pandas as pd

    pd.set_option("display.max_columns", None)
    pd.set_option("display.max_rows", None)
    pd.set_option("display.width", None)

    # Create/Recreate the Model and Endpoint
    # create_model_and_endpoint()

    # Run the tests
    test_list_inference_runs()
    test_performance_metrics()
    test_retrieval_with_capture_name()
    test_validation_predictions()
    test_inference_predictions()
    test_confusion_matrix()
    test_shap_values()
    test_metrics_with_capture_name()

    # These are longer tests (commented out for now)
    # test_auto_inference()
    # test_inference_with_capture_name("my_holdout_test")



"""Tests for Endpoint 'chaining''"""

# Workbench Imports
from workbench.api import Endpoint

end_reg = Endpoint("test-regression")
end_class = Endpoint("test-classification")


def test_reg_to_reg():
    """Test chaining two regression endpoints"""
    pred_df = end_reg.inference(end_reg.auto_inference())
    print(pred_df.columns)
    print(pred_df)


def test_reg_to_class():
    """Test chaining a regression endpoint to a classification endpoint"""
    pred_df = end_class.inference(end_reg.auto_inference())
    print(pred_df.columns)
    print(pred_df)


def test_class_to_class():
    """Test chaining two classification endpoints"""
    pred_df = end_class.inference(end_class.auto_inference())
    print(pred_df.columns)
    print(pred_df)


def test_class_to_reg():
    """Test chaining a classification endpoint to a regression endpoint"""
    pred_df = end_reg.inference(end_class.auto_inference())
    print(pred_df.columns)
    print(pred_df)


if __name__ == "__main__":

    # Set Pandas Display Options
    import pandas as pd

    pd.set_option("display.max_columns", None)
    pd.set_option("display.width", 1000)

    # Run the tests
    test_reg_to_reg()
    test_reg_to_class()
    test_class_to_class()
    test_class_to_reg()

    print("All tests passed!")



from workbench.api import FeatureSet, Model, ModelType, Endpoint

# Grab a FeatureSet
my_features = FeatureSet("aqsol_features")
model = Model("aqsol-regression")
feature_list = model.features()
target = model.target()

# Recreate Flag in case you want to recreate the artifacts
recreate = True

# PyTorch Regression Model
if recreate or not Model("aqsol-pytorch-reg").exists():
    feature_set = FeatureSet("aqsol_features")
    m = feature_set.to_model(
        name="aqsol-pytorch-reg",
        model_type=ModelType.REGRESSOR,
        model_class="PyTorch",
        feature_list=feature_list,
        target_column=target,
        description="PyTorch Regression Model for AQSol",
        tags=["pytorch", "molecular descriptors"],
        hyperparameters={"training_config": {"max_epochs": 150}, "model_config": {"layers": "128-64-32"}},
    )
    m.set_owner("BW")

# Create an Endpoint for the Regression Model
if recreate or not Endpoint("aqsol-pytorch-reg").exists():
    m = Model("aqsol-pytorch-reg")
    end = m.to_endpoint(tags=["pytorch", "molecular descriptors"])
    end.set_owner("BW")

    # Run inference on the endpoint
    end.auto_inference(capture=True)

# Pytorch Classification Model
if recreate or not Model("aqsol-pytorch-class").exists():
    feature_set = FeatureSet("aqsol_features")
    m = feature_set.to_model(
        name="aqsol-pytorch-class",
        model_type=ModelType.CLASSIFIER,
        model_class="PyTorch",
        feature_list=feature_list,
        target_column="solubility_class",
        description="PyTorch Classification Model for AQSol",
        tags=["pytorch", "molecular descriptors"],
        hyperparameters={"training_config": {"max_epochs": 150}, "model_config": {"layers": "256-128-64"}},
    )
    m.set_owner("BW")
    m.set_class_labels(["low", "medium", "high"])

# Create an Endpoint for the Classification Model
if recreate or not Endpoint("aqsol-pytorch-class").exists():
    m = Model("aqsol-pytorch-class")
    end = m.to_endpoint(tags=["pytorch", "molecular descriptors"])
    end.set_owner("BW")

    # Run inference on the endpoint
    end.auto_inference(capture=True)



import numpy as np
import matplotlib.pyplot as plt

# Example coverage probability data
nominal_confidence_levels = np.linspace(0.5, 1.0, 10)
coverage_probabilities = nominal_confidence_levels + np.random.normal(0, 0.05, 10)

# Plot
plt.figure(figsize=(10, 6))
plt.plot(
    nominal_confidence_levels,
    coverage_probabilities,
    "o-",
    label="Coverage Probability",
)
plt.plot([0.5, 1.0], [0.5, 1.0], "--", color="gray")
plt.xlabel("Nominal Confidence Level")
plt.ylabel("Coverage Probability")
plt.legend()
plt.title("Coverage Probability Plot")
plt.show()



"""FeatureSetsPageView pulls FeatureSet metadata from the AWS Service Broker with Details Panels on each FeatureSet"""

import pandas as pd

# Workbench Imports
from workbench.web_interface.page_views.page_view import PageView
from workbench.cached.cached_meta import CachedMeta
from workbench.cached.cached_feature_set import CachedFeatureSet


class FeatureSetsPageView(PageView):
    def __init__(self):
        """FeatureSetsPageView pulls FeatureSet metadata and populates a Details Panel"""
        # Call SuperClass Initialization
        super().__init__()

        # CachedMeta object for Cloud Platform Metadata
        self.meta = CachedMeta()

        # Initialize the Endpoints DataFrame
        self.feature_sets_df = None
        self.refresh()

    def refresh(self):
        """Refresh the data from the AWS Service Broker"""
        self.log.important("Calling featuresets page view refresh()..")
        self.feature_sets_df = self.meta.feature_sets(details=True)

    def feature_sets(self) -> pd.DataFrame:
        """Get a list of all the Feature

        Returns:
            pd.DataFrame: DataFrame of all the FeatureSets
        """
        return self.feature_sets_df

    @staticmethod
    def feature_set_smart_sample(feature_name: str) -> pd.DataFrame:
        """Get a smart-sample dataframe (sample + outliers) for the given FeatureSet Index
        Args:
            feature_name(str): The Name of the DataSource
        Returns:
            pd.DataFrame: The smart-sample DataFrame
        """
        fs = CachedFeatureSet(feature_name)
        if not fs.exists():
            return pd.DataFrame({"name": [feature_name], "status": ["NOT FOUND"]})
        if not fs.ready():
            status = fs.get_status()
            return pd.DataFrame({"name": [feature_name], "status": [f"{status}"]})
        else:
            # Get the display columns
            display_columns = fs.view("display").columns

            # Return the Smart Sample (with the display subset of the columns)
            smart_sample = fs.smart_sample()
            return smart_sample[[col for col in display_columns if col in smart_sample.columns]]

    @staticmethod
    def feature_set_details(feature_name: str) -> (dict, None):
        """Get all the details for the given FeatureSet Name
        Args:
            feature_name(str): The Name of the FeatureSet
        Returns:
            dict: The details for the given FeatureSet (or None if not found)
        """
        fs = CachedFeatureSet(feature_name)
        if not fs.exists() or not fs.ready():
            return None

        # Get the display columns
        display_columns = fs.view("display").columns

        # Subset some of the details using the display columns
        sub_details = fs.details()
        sub_details["column_stats"] = {
            k: sub_details["column_stats"][k] for k in display_columns if k in sub_details["column_stats"]
        }
        sub_details["column_details"] = {
            k: sub_details["column_details"][k] for k in display_columns if k in sub_details["column_details"]
        }

        # Subset the correlation details to avoid breaking the correlation matrix
        for column, data in sub_details["column_stats"].items():
            if "correlations" in data:
                data["correlations"] = {
                    k: data["correlations"][k] for k in display_columns if k in data["correlations"]
                }

        # Return the Subset of DataSource Details
        return sub_details


if __name__ == "__main__":
    # Exercising the FeatureSetsPageView
    import time
    from pprint import pprint

    # Create the class and get the AWS FeatureSet details
    feature_view = FeatureSetsPageView()

    # List the FeatureSets
    print("FeatureSetsSummary:")
    summary = feature_view.feature_sets()
    print(summary.head())

    # Get the details for the first FeatureSet
    my_feature_name = summary["Feature Group"].iloc[0]
    print("\nFeatureSetDetails:")
    details = feature_view.feature_set_details(my_feature_name)
    pprint(details)

    # Get a sample dataframe for the given FeatureSets
    print("\nSampleDataFrame:")
    sample_df = feature_view.feature_set_smart_sample(my_feature_name)
    print(sample_df.shape)
    print(sample_df.head())

    # Test refresh
    print("\nRefreshing...")
    feature_view.refresh()

    # Give any broker threads time to finish
    time.sleep(1)



"""PandasToView Class: A View that joins the source_table with a Pandas dataframe"""

from typing import Union
import pandas as pd

# Workbench Imports
from workbench.api import DataSource, FeatureSet
from workbench.core.views.view import View
from workbench.core.views.create_view import CreateView
from workbench.core.views.view_utils import dataframe_to_table, get_column_list


class PandasToView(CreateView):
    """PandasToView Class: A View that joins the source_table with a Pandas dataframe

    Common Usage:
    ```
    # Grab a DataSource
    ds = DataSource("test_data")

    # Do some awesome Feature Engineering :)
    my_df = ds.pull_dataframe()
    my_df["random1"] = np.random.rand(len(my_df))
    my_df["random2"] = np.random.rand(len(my_df))

    # Create your new View
    fe_view = PandasToView.create("feature_engineering_view", ds, df=my_df, id_column="id")

    # Query the view
    df = fe_view.query(f"SELECT * FROM {fe_view.table} where residuals > 0.5")
    ```
    """

    @classmethod
    def create(
        cls,
        view_name: str,
        artifact: Union[DataSource, FeatureSet],
        df: pd.DataFrame,
        id_column: str,
        source_table: str = None,
    ) -> Union[View, None]:
        """Factory method to create and return a PandasToView instance.

        Args:
            view_name (str): The name of the view
            artifact (Union[DataSource, FeatureSet]): The DataSource or FeatureSet object
            df (pd.DataFrame): The Pandas DataFrame to join with the source_table
            id_column (str): The name of the id column (must be defined for join logic)
            source_table (str, optional): The table/view to create the view from. Defaults to None

        Returns:
            Union[View, None]: The created View object (or None if failed to create the view)
        """
        # Instantiate the PandasToView
        instance = cls(view_name, artifact, source_table)

        # Delegate to the internal method for creating the view
        return instance._create_view(df, id_column)

    def _create_view(self, df: pd.DataFrame, id_column: str) -> Union[View, None]:
        """Internal method to create the view by combining (or joining) with a Pandas DataFrame.

        Args:
            df (pd.DataFrame): The Pandas DataFrame to join with the source_table
            id_column (str): The name of the id column (must be defined for join logic)

        Returns:
            Union[View, None]: The created View object (or None if failed to create the view)
        """
        self.log.important(f"Creating View with DF {self.table}...")

        # Check the number of rows in the source_table, if greater than 1M, then give an error and return
        row_count = self.data_source.num_rows()
        if row_count > 1_000_000:
            self.log.error(
                f"View with DF cannot be created on more than 1M rows. {self.source_table} has {row_count} rows."
            )
            return None

        # Drop any columns generated from AWS
        aws_cols = ["write_time", "api_invocation_time", "is_deleted", "event_time"]
        source_table_columns = get_column_list(self.data_source, self.source_table)
        column_list = [col for col in source_table_columns if col not in aws_cols]

        # Drop any columns generated from AWS from the incoming dataframe
        df = df.drop(columns=aws_cols, errors="ignore")

        # Enclose each column name in double quotes
        sql_columns = ", ".join([f'"{column}"' for column in column_list])

        # Pull in the data from the source_table
        query = f"SELECT {sql_columns} FROM {self.source_table}"
        source_df = self.data_source.query(query)

        # Check if the id_column exists in the source_table
        if id_column not in source_df.columns:
            self.log.error(f"id_column {id_column} not found in {self.source_table}. Cannot create the View.")
            return None

        # Check if the id_column exists in the dataframe
        if id_column not in df.columns:
            self.log.error(f"id_column {id_column} not found in the dataframe. Cannot create the View.")
            return None

        # Check for duplicate id_column values in the incoming dataframe
        if df[id_column].duplicated().any():
            self.log.error(f"id_column {id_column} has duplicate values, this will cause duplicate rows in the view!")

        # Remove any columns in the incoming df that overlap with the source_df (except for the id_column)
        overlap_columns = [col for col in df.columns if col in source_df.columns and col != id_column]
        self.log.info(f"Removing overlap columns: {overlap_columns}")
        df = df.drop(columns=overlap_columns)

        # Create a supplemental data table with the incoming dataframe
        supplemental_table = f"_{self.base_table_name}___{self.view_name}"
        dataframe_to_table(self.data_source, df, supplemental_table)

        # Create a list of columns in SQL form (for the source table)
        source_columns_str = ", ".join([f'A."{col}"' for col in source_df.columns])

        # Create a list of columns in SQL form (for the incoming dataframe table)
        df_columns_str = ", ".join([f'B."{col}"' for col in df.columns if col != id_column])

        # Construct the CREATE VIEW query
        create_view_query = f"""
        CREATE OR REPLACE VIEW "{self.table}" AS
        SELECT {source_columns_str}, {df_columns_str}
        FROM {self.source_table} A
        LEFT JOIN {supplemental_table} B
        ON A.{id_column} = B.{id_column}
        """

        # Execute the CREATE VIEW query
        self.data_source.execute_statement(create_view_query)

        # Return the View
        return View(self.data_source, self.view_name, auto_create_view=False)


if __name__ == "__main__":
    """Exercise the PandasToView functionality"""
    import numpy as np

    # Grab a FeatureSet
    fs = FeatureSet("abalone_features")

    # Generate a DataFrame with two random columns
    my_df = fs.pull_dataframe()
    my_df["random1"] = np.random.rand(len(my_df))
    my_df["random2"] = np.random.rand(len(my_df))

    # Create a PandasToView
    df_view = PandasToView.create("test_view", fs, df=my_df, id_column="auto_id")

    # Pull the dataframe view
    my_df = df_view.pull_dataframe(limit=5)
    print(my_df)

    # Delete the dataframe view
    df_view.delete()



# Workbench Imports
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp

boto3_session = AWSAccountClamp().boto3_session


# Create a Glue client
glue_client = boto3_session.client("glue")

# Define the database and table names
database_name = "workbench"
table_name = "test_table"

# Define the original column names with mixed case
original_column_dict = {"Id": "int", "Name": "string", "Age": "int"}


# Delete the table if it already exists
try:
    glue_client.delete_table(DatabaseName=database_name, Name=table_name)
except glue_client.exceptions.EntityNotFoundException:
    pass

# Create the Glue table
glue_client.create_table(
    DatabaseName=database_name,
    TableInput={
        "Name": table_name,
        "StorageDescriptor": {
            "Columns": [{"Name": name, "Type": dtype} for name, dtype in original_column_dict.items()],
            "Location": "s3://test-bucket/test_table/",
            "InputFormat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat",
            "OutputFormat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat",
            "SerdeInfo": {"SerializationLibrary": "org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe"},
        },
        "TableType": "EXTERNAL_TABLE",
    },
)

# Read back the Glue table
response = glue_client.get_table(DatabaseName=database_name, Name=table_name)

# Extract the column names from the response
retrieved_columns = [col["Name"] for col in response["Table"]["StorageDescriptor"]["Columns"]]

# Compare the original column names to the retrieved column names
original_columns = list(original_column_dict.keys())
print("\nOriginal Columns")
print(original_columns)
print("\nRetrieved Columns")
print(retrieved_columns)



"""Script that checks ModelGroups, Model (Resources), Endpoints and does a set of Sanity checks"""

import logging

# Workbench Imports
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp
from workbench.core.artifacts.model_core import ModelCore
from workbench.core.artifacts.endpoint_core import EndpointCore

# Get a sagemaker_client from the AWSAccountClamp()
sagemaker_client = AWSAccountClamp().sagemaker_client()

# Setup logging
log = logging.getLogger("workbench")
log.setLevel(logging.INFO)


def run_sanity_checks(tag: bool = False, delete_stuff: bool = False):

    # Get all the model groups
    response = sagemaker_client.list_model_package_groups(MaxResults=100)
    model_group_names = [
        model_group["ModelPackageGroupName"] for model_group in response["ModelPackageGroupSummaryList"]
    ]
    log.important(f"Found {len(model_group_names)} Model Groups")
    for model_group_name in model_group_names:
        # For each model group report the number of model packages in the group
        package_response = sagemaker_client.list_model_packages(ModelPackageGroupName=model_group_name)
        num_packages = len(package_response["ModelPackageSummaryList"])
        log.info(f"Model Group: {model_group_name} ({num_packages} packages)")

    # Get all the model packages
    all_model_packages = sagemaker_client.list_model_packages(MaxResults=100)

    # Figure out with model packages are NOT part of a model package group
    standalone_model_packages = []
    for package in all_model_packages["ModelPackageSummaryList"]:
        if "ModelPackageGroupName" in package:
            if package["ModelPackageGroupName"] not in model_group_names:
                standalone_model_packages.append(package)
        else:
            standalone_model_packages.append(package)
    log.important(f"Found {len(standalone_model_packages)} Model Packages (not in a Group)")
    for package in standalone_model_packages:
        log.important(f"\t{package['ModelPackageArn']}")

    # Get all the model resources (models not in a model group)
    response = sagemaker_client.list_models(MaxResults=100)
    model_names = [model["ModelName"] for model in response["Models"]]
    log.important(f"Found {len(model_names)} Models (not in a Model Package/Group)")
    for model_name in model_names:
        log.info(f"\tModel: {model_name}")

    # Each ModelGroup should have an endpoint and having an endpoint means
    # that a 'Model' Resource is created, so if we find a ModelGroup without
    # a Model Resource, then we have an ModelGroup without an endpoint
    model_group_set = set(model_group_names)
    model_set = set(model_names)
    model_group_without_model = []

    # Check if the model group name is a substring of any model name
    for model_group in model_group_set:
        if not any(model_group in model_name for model_name in model_set):
            model_group_without_model.append(model_group)
    if len(model_group_without_model) > 0:
        log.important(
            f"({len(model_group_without_model)}) Possible Model Groups without an Endpoint (Heuristic/substring): "
        )
        for model_group in model_group_without_model:
            log.important(f"{model_group}")
            if tag:
                m = ModelCore(model_group)
                m.add_health_tag("no_endpoint")
        log.important("Recommendation: Delete these Models Groups or create an Endpoint for them")

    # List all endpoints
    endpoints_response = sagemaker_client.list_endpoints(MaxResults=100)

    # Check each endpoint to see if it uses any of the models from the model group
    found_models = []
    log.important(f"Found {len(endpoints_response['Endpoints'])} Endpoints")
    for endpoint in endpoints_response["Endpoints"]:
        log.info(f"Endpoint: {endpoint['EndpointName']}")
        endpoint_desc = sagemaker_client.describe_endpoint(EndpointName=endpoint["EndpointName"])
        endpoint_config_desc = sagemaker_client.describe_endpoint_config(
            EndpointConfigName=endpoint_desc["EndpointConfigName"]
        )
        for variant in endpoint_config_desc["ProductionVariants"]:
            if variant["ModelName"] in model_names:
                log.info(f"\t{endpoint['EndpointName']} --> {variant['ModelName']}")
                found_models.append(variant["ModelName"])
            else:
                log.warning(f"\t{endpoint['EndpointName']} --> Model not found: {variant['ModelName']}")
                log.warning("Recommendation: This endpoint may no longer work, test it!")
                if tag:
                    e = EndpointCore(endpoint["EndpointName"])
                    e.add_health_tag("no_model")

    # Now report on the models that are not used by any endpoint (orphans)
    unused_models = model_set - set(found_models)
    log.important(f"Found {len(unused_models)} Model Resources without an Endpoint (Orphans) ")
    for model in unused_models:
        log.important(f"\t{model}")
    log.important("Recommendation: Delete these Models")
    if delete_stuff:
        for model in unused_models:
            delete_model(model, sagemaker_client)
            ModelCore.managed_delete(model)


def delete_model(model_name: str, sm_client):
    log.important(f"Deleting Model '{model_name}'...")
    sm_client.delete_model(ModelName=model_name)


if __name__ == "__main__":
    import argparse

    # We're going to have a --verbose flag and that's it
    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", action="store_true")
    parser.add_argument("--delete-stuff", action="store_true")
    args, commands = parser.parse_known_args()

    # Call the sanity checks
    run_sanity_checks(tag=args.tag, delete_stuff=args.delete_stuff)



import boto3
import time
import awswrangler as wr

# Workbench imports
from workbench.utils.performance_utils import performance


@performance
def compute_athena_table_hash(database: str, table: str, session: boto3.session, s3_scratch: str) -> str:
    """
    Compute a hash of an Athena table by concatenating all rows and columns.

    Args:
        database (str): The database name
        table (str): The table name
        session (boto3.session.Session): The boto3 session
        s3_scratch (str): S3 bucket and prefix for Athena query results
    Returns:
        str: MD5 hash of the table content
    """
    athena = session.client("athena")
    s3 = session.client("s3")

    def check_query_status(execution_id):
        """Check the status of an Athena query and return detailed failure info if applicable."""
        query_execution = athena.get_query_execution(QueryExecutionId=execution_id)
        status = query_execution["QueryExecution"]["Status"]
        state = status["State"]

        if state == "FAILED":
            reason = status.get("StateChangeReason", "No additional details provided.")
            print(f"Query failed. Reason: {reason}")
            raise RuntimeError(f"Athena query failed with reason: {reason}")
        elif state == "CANCELLED":
            print("Query was cancelled.")
            raise RuntimeError("Athena query was cancelled.")
        return state

    # Step 1: Retrieve the schema for the table
    schema_query = f"""
    SELECT column_name
    FROM information_schema.columns
    WHERE table_schema = '{database}' AND table_name = '{table}'
    ORDER BY ordinal_position
    """

    schema_execution_id = athena.start_query_execution(
        QueryString=schema_query,
        QueryExecutionContext={"Database": "information_schema"},
        ResultConfiguration={"OutputLocation": s3_scratch},
    )["QueryExecutionId"]

    # Wait for schema query to complete
    while True:
        state = check_query_status(schema_execution_id)
        if state in ["SUCCEEDED", "FAILED", "CANCELLED"]:
            break
        time.sleep(1)

    # Retrieve schema query results
    schema_result = athena.get_query_results(QueryExecutionId=schema_execution_id)
    columns = [row["Data"][0]["VarCharValue"] for row in schema_result["ResultSet"]["Rows"][1:]]  # Skip header

    # Step 2: Build the hash query dynamically
    column_concat = ", '|', ".join([f"CAST({col} AS VARCHAR)" for col in columns])
    hash_query = f"""
    SELECT MD5(
        CAST(
            array_join(
                array_agg(
                    CONCAT({column_concat})
                ),
                '|'
            ) AS varbinary
        )
    ) AS table_hash
    FROM {database}.{table}
    """

    hash_execution_id = athena.start_query_execution(
        QueryString=hash_query,
        QueryExecutionContext={"Database": database},
        ResultConfiguration={"OutputLocation": s3_scratch},
    )["QueryExecutionId"]

    # Wait for hash query to complete
    while True:
        state = check_query_status(hash_execution_id)
        if state in ["SUCCEEDED", "FAILED", "CANCELLED"]:
            break
        time.sleep(1)

    # Retrieve the S3 location of the hash query result
    query_execution = athena.get_query_execution(QueryExecutionId=hash_execution_id)
    result_location = query_execution["QueryExecution"]["ResultConfiguration"]["OutputLocation"]
    bucket, key = result_location.replace("s3://", "").split("/", 1)

    # Retrieve hash query result from S3
    obj = s3.get_object(Bucket=bucket, Key=key)
    table_hash = obj["Body"].read().decode("utf-8").splitlines()[1]  # Skip header row
    table_hash = table_hash.replace(" ", "").replace('"', "")  # Remove spaces and quotes from the hash

    # Clean up the temporary result files
    wr.s3.delete_objects(path=s3_scratch)

    return table_hash


if __name__ == "__main__":
    # Test the athena utils functions
    from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp
    from workbench.utils.config_manager import ConfigManager

    # Get our boto3 session
    session = AWSAccountClamp().boto3_session

    # Get our workbench bucket
    cm = ConfigManager()
    workbench_bucket = cm.get_config("WORKBENCH_BUCKET")

    # Setup a temporary S3 prefix for the Athena output
    s3_scratch = f"s3://{workbench_bucket}/temp/athena_output"

    # Compute the hash for a table
    database_name = "workbench"
    table_name = "abalone_data"
    table_hash = compute_athena_table_hash(database_name, table_name, session, s3_scratch)
    print(f"Hash for {database_name}.{table_name}: {table_hash}")

    # Remove the temporary S3 prefix
    s3 = session.client("s3")
    s3.delete_object(Bucket=workbench_bucket, Key=f"temp/athena_output/{table_hash}.csv")



"""Main: The main Workbench Web Interface to view, interact, and manage Workbench Artifacts"""

from dash import register_page

# Workbench Imports
from workbench.web_interface.page_views.main_page import MainPage
from workbench.web_interface.components.plugins.ag_table import AGTable

# Local Imports
from .layout import main_layout
from . import callbacks

register_page(
    __name__,
    path="/",
    name="Workbench",
)

# Create a table object for each Workbench Artifact Type
tables = dict()
table_names = ["data_sources", "feature_sets", "models", "endpoints"]
for table_name in table_names:
    tables[table_name] = AGTable()

# Create a table container for each Workbench Artifact Type
table_containers = dict()
table_name = "data_sources"
table_containers[table_name] = tables[table_name].create_component(
    f"main_{table_name}",
    header_color="rgb(120, 70, 70)",
    max_height=200,
)
table_name = "feature_sets"
table_containers[table_name] = tables[table_name].create_component(
    f"main_{table_name}",
    header_color="rgb(110, 110, 70)",
    max_height=200,
)
table_name = "models"
table_containers[table_name] = tables[table_name].create_component(
    f"main_{table_name}",
    header_color="rgb(60, 100, 60)",
    max_height=500,
)
table_name = "endpoints"
table_containers[table_name] = tables[table_name].create_component(
    f"main_{table_name}",
    header_color="rgb(100, 60, 100)",
    max_height=500,
)

# Set up our layout (Dash looks for a var called layout)
layout = main_layout(**table_containers)

# Grab a view that gives us a summary of all the artifacts currently in Workbench
main_page_view = MainPage()

# Set up the callbacks for tables on the main page
callbacks.last_updated()
callbacks.plugin_page_info()
callbacks.tables_refresh(main_page_view, tables)

# Set up our subpage navigation
callbacks.navigate_to_subpage()



"""Main Layout: Layout for the Main page in the Artifact Viewer"""

from dash import html, dcc
import dash_bootstrap_components as dbc

# Workbench Imports
import workbench
from workbench.web_interface.components.plugins.ag_table import AGTable
from workbench.utils.theme_manager import ThemeManager
from workbench.utils.config_manager import ConfigManager

# Get the UI update rate
update_rate = ConfigManager().ui_update_rate() * 1000  # Convert to Milliseconds

# Set the Main Page Title
tm = ThemeManager()
app_title = tm.branding().get("app_title", "Workbench Dashboard")


def main_layout(
    data_sources: AGTable,
    feature_sets: AGTable,
    models: AGTable,
    endpoints: AGTable,
) -> html.Div:
    """Main Layout for the Dashboard"""
    workbench_version = workbench.__version__.split("+")[0].strip()
    cm = ConfigManager()
    license_id = cm.get_license_id()

    # Define the layout with one table per row
    layout = html.Div(
        children=[
            # This refreshes the page at the specified intervals
            dcc.Interval(id="main_page_refresh", interval=update_rate, n_intervals=0),
            dcc.Store(
                id="table_hashes",
                data={
                    "data_sources": None,
                    "feature_sets": None,
                    "models": None,
                    "endpoints": None,
                },
            ),
            # Top of Main Page Header/Info Section
            dbc.Row(
                [
                    html.Div(
                        [
                            html.H2(
                                [
                                    html.A(
                                        app_title + " ",
                                        href="/status",
                                        style={
                                            "textDecoration": "none",
                                            "color": "inherit",
                                        },
                                    ),
                                    html.Span(
                                        f"{workbench_version}",
                                        className="orange-text",
                                        style={"fontSize": 15},
                                    ),
                                    html.A(
                                        f"  [{license_id}]",
                                        href="/license",
                                        className="pink-text",
                                        style={
                                            "fontSize": 15,
                                            "textDecoration": "none",
                                        },
                                    ),
                                ],
                                style={"marginBottom": "0px"},
                            ),
                            html.Div(
                                "Last Updated: ",
                                id="data-last-updated",
                                className="green-text",
                                style={
                                    "fontSize": 15,
                                    "fontStyle": "italic",
                                    "fontWeight": "bold",
                                    "paddingLeft": "200px",
                                },
                            ),
                        ]
                    )
                ]
            ),
            # Plugin Page Links
            dbc.Row(
                html.Div(
                    [
                        html.H4("Plugin Pages", id="plugin-pages-header", style={"textAlign": "left"}),
                        html.Ul([], id="plugin-pages-list"),  # Placeholder for the list
                    ],
                    id="plugin-pages",
                ),
                style={"padding": "20px 0px"},
            ),
            # Each table in its own row
            dbc.Row(
                [
                    html.H3("Data Sources", style={"textAlign": "left"}),
                    data_sources,
                ],
                style={"padding": "0px 0px"},
            ),
            dbc.Row(
                [
                    html.H3("Feature Sets", style={"textAlign": "left"}),
                    feature_sets,
                ],
                style={"padding": "20px 0px"},
            ),
            dbc.Row(
                [
                    html.H3("Models", style={"textAlign": "left"}),
                    models,
                ],
                style={"padding": "20px 0px"},
            ),
            dbc.Row(
                [
                    html.H3("Endpoints", style={"textAlign": "left"}),
                    endpoints,
                ],
                style={"padding": "20px 0px"},
            ),
        ],
        style={"margin": "30px"},
    )
    return layout



"""Markdown Utility/helper methods"""

from workbench.utils.symbols import health_icons


def health_tag_markdown(health_tags: list[str]) -> str:
    """Internal method to generate the health tag markdown
    Args:
        health_tags (list[str]): A list of health tags
    Returns:
        str: A markdown string
    """
    # If we have no health tags, then add a bullet for healthy
    markdown = "**Health Checks**\n"  # Header for Health Checks

    # If we have no health tags, then add a bullet for healthy
    if not health_tags:
        markdown += f"* Healthy: {health_icons.get('healthy')}\n\n"
        return markdown

    # Special case for no_activity with no other tags
    if len(health_tags) == 1 and health_tags[0] == "no_activity":
        markdown += f"* Healthy: {health_icons.get('healthy')}\n"
        markdown += f"* No Activity: {health_icons.get('no_activity')}\n\n"
        return markdown

    # If we have health tags, then add a bullet for each tag
    markdown += "\n".join(f"* {tag}: {health_icons.get(tag, '')}" for tag in health_tags)
    markdown += "\n\n"
    return markdown


def tag_styling(tags: list, tag_dict: dict) -> str:
    """Generate a Markdown string with styled spans for tags.

    Args:
        tags (list): List of tags to style.
        tag_dict (dict): Dictionary mapping substrings to CSS classes.

    Returns:
        str: Markdown string with styled spans.
    """
    styled_tags = []
    for tag in tags:
        # Check for a matching class in the dictionary
        class_name = next((class_name for substring, class_name in tag_dict.items() if substring in tag.lower()), None)
        if class_name:
            # Add span with the matched class
            styled_tags.append(f'<span class="{class_name}" style="padding:0" children="{tag}"/>')
        else:
            # Add plain tag if no class matches
            styled_tags.append(tag)
    # Join styled tags with commas
    return ", ".join(styled_tags)


def tags_to_markdown(tags: str) -> str:
    """Convert tags to a Markdown string.

    Args:
        tags (str): Deliminator-separated string of tags.

    Returns:
        str: Markdown string with tags.
    """
    # Split the tags (this needs to be refactored to use a list)
    tag_list = tags.split("::")

    # Separate items with ":" from those without
    with_colon = [item for item in tag_list if ":" in item]
    without_colon = [item for item in tag_list if ":" not in item]
    ordered_tag_list = with_colon
    if without_colon:
        without_colon = ", ".join(without_colon)
        ordered_tag_list += [without_colon]
    tag_markdown = "**Tags:**\n"
    markdown_tag_list = []
    for tag in ordered_tag_list:
        if ":" in tag:
            markdown_tag_list.append(f"*{tag.split(':')[0]}:*{tag.split(':')[1]}")
        else:
            markdown_tag_list.append(tag)
    tag_markdown += ", ".join(markdown_tag_list)
    return tag_markdown


def dict_to_markdown(data: dict, title: str = None) -> str:
    """Convert a dictionary to pretty Markdown format.
    Args:
        data (dict): Dictionary to convert
        title (str, optional): Optional title for the dictionary
    Returns:
        str: Markdown formatted string
    """

    def _convert_dict(data: dict, indent_level: int = 0) -> str:
        markdown = ""
        indent_str = "    " * indent_level  # 4 spaces per level

        for key, value in data.items():
            if isinstance(value, dict):
                # Nested dictionary
                markdown += f"{indent_str}* *{key}:*\n"
                markdown += _convert_dict(value, indent_level + 1)
            elif isinstance(value, list):
                # Check if it's a list of dictionaries
                if value and all(isinstance(item, dict) for item in value):
                    # List of dictionaries
                    markdown += f"{indent_str}* *{key}:*\n"
                    for i, dict_item in enumerate(value):
                        markdown += f"{indent_str}    * Item {i + 1}:\n"
                        markdown += _convert_dict(dict_item, indent_level + 2)
                else:
                    # Regular list values
                    markdown += f"{indent_str}* *{key}:* {', '.join(map(str, value))}\n"
            else:
                # Simple key-value pair
                markdown += f"{indent_str}* *{key}:* {value}\n"

        return markdown

    # Add title if provided (always at root level, no indentation)
    result = ""
    if title:
        result += f"**{title}**\n\n"
    result += _convert_dict(data)
    return result


def dict_to_collapsible_html(data: dict, title: str = None, collapse_all: bool = False) -> str:
    """Convert a dictionary to collapsible HTML format.
    Args:
        data (dict): Dictionary to convert
        title (str, optional): Optional title for the dictionary
        collapse_all (bool): Whether to collapse all sections by default. Defaults to False.
    Returns:
        str: HTML formatted string with collapsible sections
    """

    def _convert_dict_html(data: dict, indent_level: int = 0) -> str:
        html = ""
        indent_style = f'style="margin-left: {indent_level * 20}px;"' if indent_level > 0 else ""
        # Leaf nodes get slightly less indentation to align with content
        leaf_indent_style = (
            f'style="margin-left: {max(0, indent_level - 1) * 10 + 10}px;"'
            if indent_level > 0
            else 'style="margin-left: 10px;"'
        )

        # Determine if details should be open or closed
        open_attr = "" if collapse_all else "open"

        for key, value in data.items():
            if isinstance(value, dict):
                html += f"<details {open_attr} {indent_style}><summary><b>{key}</b></summary>\n"
                html += _convert_dict_html(value, indent_level + 1)
                html += "</details>\n"
            elif isinstance(value, list):
                if value and all(isinstance(item, dict) for item in value):
                    html += f"<details {open_attr} {indent_style}><summary><b>{key}</b></summary>\n"
                    for i, dict_item in enumerate(value):
                        html += f'<details {open_attr} style="margin-left: {(indent_level + 1) * 20}px;">'
                        html += f"<summary>Item {i + 1}</summary>\n"
                        html += _convert_dict_html(dict_item, indent_level + 2)
                        html += "</details>\n"
                    html += "</details>\n"
                else:
                    # Leaf node - use bullet with reduced indentation
                    html += f"<div {leaf_indent_style}>• <em>{key}:</em> {', '.join(map(str, value))}</div>\n"
            else:
                # Leaf node - use bullet with reduced indentation
                html += f"<div {leaf_indent_style}>• <em>{key}:</em> {value}</div>\n"
        return html

    # Add title and content
    result = ""
    open_attr = "" if collapse_all else "open"

    if title:
        result += f"<details {open_attr}><summary><strong>{title}</strong></summary>\n"
        result += _convert_dict_html(data, 1)  # Start with indent level 1 under title
        result += "</details>\n"
    else:
        result += _convert_dict_html(data, 0)  # Start with no indent
    return result


if __name__ == "__main__":
    """Exercise the Markdown Utilities"""
    from workbench.api.model import Model

    # Open a model
    model = Model("abalone-regression")
    health_tags = model.get_health_tags()

    # Print the health tag markdown
    print(health_tag_markdown(health_tags))

    # Print the tag markdown
    tag_str = "tag1::tag2::tag3::key1:value1::key2:value2"
    print(tags_to_markdown(tag_str))

    # Print the tag markdown
    tag_str = "key1:value1::key2:value2::key3:value3"
    print(tags_to_markdown(tag_str))

    # Print the dict to markdown
    sample_dict = {
        "key1": "value1",
        "key2": {"subkey1": "subvalue1", "subkey2": ["item1", "item2"]},
        "key3": ["list_item1", "list_item2"],
        "key4": [{"nested_key": "nested_value"}, {"another_key": "another_value"}],
    }
    print(dict_to_markdown(sample_dict, title="Sample Dictionary"))
    print(dict_to_collapsible_html(sample_dict, title="Sample Dictionary"))



"""License:  A Workbench Web Interface to view License Details"""

from dash import register_page
from dash import html

# Workbench Imports
from workbench.web_interface.components.plugins import license_details
from workbench.utils.config_manager import ConfigManager
from workbench.utils.license_manager import LicenseManager

# Register this page with Dash
register_page(
    __name__,
    path="/license",
    name="Workbench - License",
)

# Grab the API Key from the Workbench ConfigManager
cm = ConfigManager()
api_key = cm.get_config("WORKBENCH_API_KEY")
license_api_key = cm.get_config("LICENSE_API_KEY")
my_license_info = LicenseManager.load_api_license(aws_account_id=None, api_key=api_key, license_api_key=license_api_key)

# Create a Markdown component to display the license details
markdown_details = license_details.LicenseDetails()
details_component = markdown_details.create_component("license_details")
updated_properties = markdown_details.update_properties(my_license_info)

# Set the properties directly on the server side before initializing the layout
for (component_id, prop), value in zip(markdown_details.properties, updated_properties):
    for child in details_component.children:
        if child.id == component_id:
            setattr(child, prop, value)
            break

# Simple layout for the license details
layout = html.Div(
    children=[
        html.H2("Workbench: License"),
        details_component,
    ],
    style={"padding": "12px 30px 30px 30px"},
)



"""Example for computing Canonicalize SMILES strings"""

import pandas as pd
from workbench.utils.chem_utils import canonicalize

test_data = [
    {"id": "Acetylacetone", "smiles": "CC(=O)CC(=O)C", "expected": "CC(=O)CC(C)=O"},
    {"id": "Imidazole", "smiles": "c1cnc[nH]1", "expected": "c1c[nH]cn1"},
    {"id": "Pyridone", "smiles": "C1=CC=NC(=O)C=C1", "expected": "O=c1cccccn1"},
    {"id": "Guanidine", "smiles": "C(=N)N=C(N)N", "expected": "N=CN=C(N)N"},
    {"id": "Catechol", "smiles": "c1cc(c(cc1)O)O", "expected": "Oc1ccccc1O"},
    {"id": "Formamide", "smiles": "C(=O)N", "expected": "NC=O"},
    {"id": "Urea", "smiles": "C(=O)(N)N", "expected": "NC(N)=O"},
    {"id": "Phenol", "smiles": "c1ccc(cc1)O", "expected": "Oc1ccccc1"},
]

# Convert test data to a DataFrame
df = pd.DataFrame(test_data)

# Perform canonicalization
result_df = canonicalize(df)
print(result_df)



"""PipelineExecutor: Internal Class: Executes a Workbench Pipeline"""

import logging

# Workbench Imports
from workbench.api import DataSource, FeatureSet, Model, Endpoint
from workbench.api.model import ModelType


class PipelineExecutor:
    """PipelineExecutor: Internal Class: Executes a Workbench Pipeline

    Common Usage:
        ```python
        my_pipeline = PipelineExecutor(pipeline)
        my_pipeline.execute()  # Execute entire pipeline
        my_pipeline.execute_partial(["data_source", "feature_set"])
        my_pipeline.execute_partial(["model", "endpoint"])
        ```
    """

    def __init__(self, pipeline):
        """PipelineExecutor Init Method"""
        self.log = logging.getLogger("workbench")
        self.pipeline_name = pipeline.name
        self.pipeline = pipeline.pipeline

    def execute(self, subset: list = None):
        """Execute the Workbench Pipeline

        Args:
            subset (list): A list of steps to execute. If None, execute the entire pipeline

        Raises:
            RuntimeError: If the pipeline execution fails in any way
        """
        self.log.important(f"Executing Pipeline {self.pipeline_name}...")
        if subset:
            self.log.important(f"\tSubset: {subset}")
        workbench_objects = {}
        for class_name, kwargs in self.pipeline.items():

            # Input is a special case
            data_input = kwargs["input"]
            del kwargs["input"]
            if isinstance(input, str) and data_input == "<<parameter_required>>":
                msg = "Call set_input() to set the input (DataFrame, file path or S3 path)"
                self.log.critical(msg)
                raise RuntimeError(msg)

            # DataSource
            if class_name == "data_source":
                # Don't create a DataSource if we're only executing a subset
                if not subset or "data_source" in subset:
                    workbench_objects["data_source"] = DataSource(data_input, **kwargs)
                else:
                    workbench_objects["data_source"] = DataSource(source=kwargs["name"])

            # FeatureSet
            elif class_name == "feature_set":

                # Special case for holdout_ids
                holdout_ids = None
                if "holdout_ids" in kwargs:
                    if kwargs["holdout_ids"] == "<<parameter_optional>>":
                        self.log.important("Hold out ids are not set, defaulting to 80/20 split")
                        holdout_ids = None
                        id_column = None
                    else:
                        holdout_ids = kwargs["holdout_ids"]
                        if "id_column" not in kwargs or (kwargs["id_column"] == "<<parameter_optional>>"):
                            self.log.warning(
                                "Hold out ids are set, but no id column is provided! Defaulting to 80/20 split"
                            )
                            holdout_ids = None
                            id_column = None
                        else:
                            id_column = kwargs["id_column"]
                            del kwargs["id_column"]
                    del kwargs["holdout_ids"]

                # Check for a transform and create a FeatureSet
                if "data_source" in workbench_objects and (not subset or "feature_set" in subset):
                    # Special case for feature_schema
                    if "feature_schema" in kwargs:
                        # Hard coded: Cheese sauce for now
                        if kwargs["feature_schema"] == "molecular_descriptors_v1":
                            del kwargs["feature_schema"]
                            self.log.error("Feature Schema: molecular_descriptors_v1 not currently implemented")
                            # rdkit_features = MolecularDescriptors(data_input, kwargs["name"])
                            # rdkit_features.set_output_tags(kwargs["tags"])
                            # query = f"SELECT id, solubility, smiles FROM {data_input}"
                            # rdkit_features.transform(id_column=kwargs["id_column"], query=query)
                            # rdkit_features.transform(id_column=kwargs["id_column"])
                        else:
                            raise RuntimeError(f"Unsupported feature schema: {kwargs['feature_schema']}")
                    else:
                        # Normal case for feature_set
                        workbench_objects["data_source"].to_features(**kwargs)

                # Create the FeatureSet and set hold out ids if specified
                workbench_objects["feature_set"] = FeatureSet(kwargs["name"])
                if holdout_ids:
                    workbench_objects["feature_set"].set_training_holdouts(id_column, holdout_ids)

            # Model
            elif class_name == "model":

                # Special case for model type
                if "model_type" in kwargs:
                    kwargs["model_type"] = ModelType(kwargs["model_type"])

                # Check for a transform and create a Model
                if "feature_set" in workbench_objects and (not subset or "model" in subset):
                    workbench_objects["feature_set"].to_model(**kwargs)
                if not subset or "endpoint" in subset:
                    workbench_objects["model"] = Model(kwargs["name"])
                    workbench_objects["model"].set_owner("pipeline")

            # Endpoint
            elif class_name == "endpoint":
                # Check for a transform
                if "model" in workbench_objects and (not subset or "endpoint" in subset):
                    workbench_objects["model"].to_endpoint(**kwargs)
                    endpoint = Endpoint(kwargs["name"])
                    endpoint.auto_inference(capture=True)

            # Found something weird
            else:
                raise RuntimeError(f"Unsupported pipeline stage: {class_name}")

    def execute_partial(self, subset: list):
        """Execute a partial Pipeline

        Args:
            subset (list): A subset of the pipeline to execute

        Raises:
            RunTimeException: If the pipeline execution fails in any way
        """
        self.execute(subset)


if __name__ == "__main__":
    """Exercise the PipelineExecutor Class"""
    from workbench.api.pipeline import Pipeline

    # Retrieve an existing Pipeline
    pipeline = Pipeline("abalone_pipeline_v1")

    # Create a PipelineExecutor
    pipeline_executor = PipelineExecutor(pipeline)

    # Run Tests
    # Execute the PipelineExecutor
    # pipeline_executor.execute()

    # Execute partial Pipelines
    # pipeline_executor.execute_partial(["data_source"])
    # pipeline_executor.execute_partial(["data_source", "feature_set"])
    # pipeline_executor.execute_partial(["model", "endpoint"])



"""This Script creates realtime AQSol (Public) Artifacts in AWS/Workbench


Endpoints:
    - tautomerize-v0-rt
    - tautomerize-v0-rt-fast1
    - smiles-to-taut-md-stereo-v0-rt
    - aqsol-mol-class-rt
"""

import logging

# Workbench Imports
from workbench.api import Model, Endpoint


log = logging.getLogger("workbench")


if __name__ == "__main__":

    # Recreate Flag in case you want to recreate the artifacts
    recreate = False

    # Create the realtime Tautomerize Endpoint
    if recreate or not Endpoint("tautomerize-v0-rt").exists():
        m = Model("tautomerize-v0")
        m.set_owner("BW")
        end = m.to_endpoint(name="tautomerize-v0-rt", tags=["smiles", "tautomerization", "realtime"], serverless=False)

        # Run inference on the endpoint
        end.auto_inference(capture=True)

    # Create realtime endpoint for testing with a ml.c7i.large instance
    if recreate or not Endpoint("tautomerize-v0-rt-fast1").exists():
        m = Model("tautomerize-v0")
        m.set_owner("BW")
        end = m.to_endpoint(
            name="tautomerize-v0-rt-fast1",
            tags=["smiles", "tautomerization", "realtime"],
            serverless=False,
            instance="ml.c7i.large",
        )

        # Run inference on the endpoint
        end.auto_inference(capture=True)

    # Create realtime endpoint for testing with a ml.c7i.xlarge instance
    if recreate or not Endpoint("tautomerize-v0-rt-fast2").exists():
        m = Model("tautomerize-v0")
        m.set_owner("BW")
        end = m.to_endpoint(
            name="tautomerize-v0-rt-fast2",
            tags=["smiles", "tautomerization", "realtime"],
            serverless=False,
            instance="ml.c7i.xlarge",
        )

        # Run inference on the endpoint
        end.auto_inference(capture=True)

    # Create a realtime endpoint for Molecular Descriptors Transformer Model
    if recreate or not Endpoint("smiles-to-taut-md-stereo-v0-rt").exists():
        m = Model("smiles-to-taut-md-stereo-v0")
        m.set_owner("BW")
        end = m.to_endpoint(
            name="smiles-to-taut-md-stereo-v0-rt",
            tags=["smiles", "molecular descriptors", "realtime"],
            serverless=False,
        )

        # Run inference on the endpoint
        end.auto_inference(capture=True)

    # Create a realtime endpoint for AQSOL solubility classification model
    if recreate or not Endpoint("aqsol-mol-class-rt").exists():
        m = Model("aqsol-mol-class")
        m.set_owner("BW")
        end = m.to_endpoint(name="aqsol-mol-class-rt", tags=["aqsol", "class", "realtime"], serverless=False)



from workbench.api.feature_set import FeatureSet
from workbench.api.model import Model
from workbench.api.endpoint import Endpoint

# Grab an existing Endpoint
endpoint = Endpoint("abalone-regression")

# Workbench has full ML Pipeline provenance, so we can backtrack the inputs,
# get a DataFrame of data (not used for training) and run inference
model = Model(endpoint.get_input())
fs = FeatureSet(model.get_input())
athena_table = fs.view("training").table
df = fs.query(f"SELECT * FROM {athena_table} where training = FALSE")

# Run inference/predictions on the Endpoint
results_df = endpoint.inference(df)

# Run inference/predictions and capture the results
results_df = endpoint.inference(df, capture_name="test_inference")

# Run inference/predictions using the FeatureSet evaluation data
results_df = endpoint.auto_inference(capture=True)



import base64
import pandas as pd
from dash import dcc, html, callback, Input, Output, no_update
import plotly.graph_objects as go
import plotly.express as px
from dash.exceptions import PreventUpdate

# Workbench Imports
from workbench.web_interface.components.plugin_interface import PluginInterface, PluginPage, PluginInputType
from workbench.utils.theme_manager import ThemeManager
from workbench.utils.plot_utils import prediction_intervals


class ScatterPlot(PluginInterface):
    """A Scatter Plot Plugin for Feature Sets."""

    # Initialize this Plugin Component Class with required attributes
    auto_load_page = PluginPage.NONE
    plugin_input_type = PluginInputType.DATAFRAME

    def __init__(self, show_axes: bool = True):
        """Initialize the Scatter Plot Plugin

        Args:
            show_axes (bool): Whether to show the axes and grid. Default is True.
        """
        self.component_id = None
        self.hover_columns = []
        self.df = None
        self.show_axes = show_axes
        self.theme_manager = ThemeManager()
        self.colorscale = self.theme_manager.colorscale()

        # Call the parent class constructor
        super().__init__()

    def create_component(self, component_id: str) -> html.Div:
        """Create a Dash Graph Component without any data.

        Args:
            component_id (str): The ID of the web component.

        Returns:
            html.Div: A Dash Div Component containing the graph and dropdowns.
        """
        self.component_id = component_id

        # Fill in plugin properties and signals
        self.properties = [
            (f"{component_id}-graph", "figure"),
            (f"{component_id}-x-dropdown", "options"),
            (f"{component_id}-y-dropdown", "options"),
            (f"{component_id}-color-dropdown", "options"),
            (f"{component_id}-label-dropdown", "options"),
            (f"{component_id}-x-dropdown", "value"),
            (f"{component_id}-y-dropdown", "value"),
            (f"{component_id}-color-dropdown", "value"),
        ]
        self.signals = [(f"{component_id}-graph", "hoverData"), (f"{component_id}-graph", "clickData")]

        # Create the Composite Component
        # - A Graph/ScatterPlot Component
        # - Dropdowns for X, Y, Color, and Label
        # - Checkbox for Regression Line
        return html.Div(
            children=[
                # Main Scatter Plot Graph
                dcc.Graph(
                    id=f"{component_id}-graph",
                    figure=self.display_text("Waiting for Data..."),
                    config={"scrollZoom": True},
                    style={"height": "100%"},
                    clear_on_unhover=True,
                ),
                # Controls: X, Y, Color, Label Dropdowns, and Regression Line Checkbox
                html.Div(
                    [
                        html.Label("X", style={"marginLeft": "40px", "marginRight": "5px", "fontWeight": "bold"}),
                        dcc.Dropdown(
                            id=f"{component_id}-x-dropdown",
                            className="dropdown",
                            style={"min-width": "50px", "flex": 1},  # Responsive width
                            clearable=False,
                        ),
                        html.Label("Y", style={"marginLeft": "30px", "marginRight": "5px", "fontWeight": "bold"}),
                        dcc.Dropdown(
                            id=f"{component_id}-y-dropdown",
                            className="dropdown",
                            style={"min-width": "50px", "flex": 1},  # Responsive width
                            clearable=False,
                        ),
                        html.Label("Color", style={"marginLeft": "30px", "marginRight": "5px", "fontWeight": "bold"}),
                        dcc.Dropdown(
                            id=f"{component_id}-color-dropdown",
                            className="dropdown",
                            style={"min-width": "50px", "flex": 1},  # Responsive width
                            clearable=False,
                        ),
                        html.Label("Label", style={"marginLeft": "30px", "marginRight": "5px", "fontWeight": "bold"}),
                        dcc.Dropdown(
                            id=f"{component_id}-label-dropdown",
                            className="dropdown",
                            style={"min-width": "50px", "flex": 1},
                            options=[{"label": "None", "value": "none"}],
                            value="none",
                            clearable=False,
                        ),
                        dcc.Checklist(
                            id=f"{component_id}-regression-line",
                            options=[{"label": " Diagonal", "value": "show"}],
                            value=[],
                            style={"margin": "10px"},
                        ),
                    ],
                    style={"padding": "0px 0px 10px 0px", "display": "flex", "gap": "10px"},
                ),
                dcc.Tooltip(
                    id=f"{component_id}-overlay",
                    background_color="rgba(0,0,0,0)",
                    border_color="rgba(0,0,0,0)",
                    direction="bottom",
                    loading_text="",
                ),
            ],
            style={"height": "100%", "display": "flex", "flexDirection": "column"},  # Full viewport height
        )

    def update_properties(self, input_data: pd.DataFrame, **kwargs) -> list:
        """Update the property values for the plugin component.

        Args:
            input_data (pd.DataFrame): The input data object.
            **kwargs: Additional keyword arguments (plugins can define their own arguments).
                      Note: The current kwargs processed are:
                            - x: The default x-axis column
                            - y: The default y-axis column
                            - color: The default color column
                            - dropdown_columns: The columns to use for the x and y options
                            - hover_columns: The columns to show when hovering over a point
                            - suppress_hover_display: Suppress hover display (default: False)
                            - custom_data: Custom data that get passed to hoverData callbacks

        Returns:
            list: A list of updated property values (figure, x options, y options, color options,
                                                    label options, x default, y default,
                                                    color default).
        """
        # Get the limit for the number of rows to plot
        limit = kwargs.get("limit", 20000)

        # Ensure input_data is a DataFrame and sample if necessary
        if isinstance(input_data, pd.DataFrame):
            self.df = input_data.sample(n=limit) if len(input_data) > limit else input_data
        else:
            raise ValueError("The input data must be a Pandas DataFrame.")

        # Remove AWS created columns
        aws_cols = ["write_time", "api_invocation_time", "is_deleted", "event_time"]
        self.df = self.df.drop(columns=aws_cols, errors="ignore")

        # Set hover columns and custom data
        self.hover_columns = kwargs.get("hover_columns", self.df.columns.tolist()[:10])
        self.suppress_hover_display = kwargs.get("suppress_hover_display", False)
        self.custom_data = kwargs.get("custom_data", [])

        # Identify numeric columns
        numeric_columns = self.df.select_dtypes(include="number").columns.tolist()
        if len(numeric_columns) < 3:
            raise ValueError("At least three numeric columns are required for x, y, and color.")

        # Default x, y, and color (for color, default to a numeric column)
        x_default = kwargs.get("x", numeric_columns[0])
        y_default = kwargs.get("y", numeric_columns[1])
        color_default = kwargs.get("color", numeric_columns[2])
        regression_line = kwargs.get("regression_line", False)

        # Create the default scatter plot
        figure = self.create_scatter_plot(self.df, x_default, y_default, color_default, "none", regression_line)

        # Dropdown options for x and y: use provided dropdown_columns or fallback to numeric columns
        dropdown_columns = kwargs.get("dropdown_columns", numeric_columns)
        x_options = [{"label": col, "value": col} for col in dropdown_columns]
        y_options = x_options.copy()

        # For color dropdown include any categorical columns (with less than 20 unique values)
        cat_columns = self.df.select_dtypes(include=["object", "string", "category"]).columns.tolist()
        cat_columns = [col for col in cat_columns if self.df[col].astype(str).nunique() < 20]
        color_columns = numeric_columns + cat_columns
        color_options = [{"label": col, "value": col} for col in color_columns]

        # For label dropdown, include None option and all columns
        label_options = [{"label": "None", "value": "none"}]
        label_options.extend([{"label": col, "value": col} for col in self.df.columns])

        return [figure, x_options, y_options, color_options, label_options, x_default, y_default, color_default]

    def create_scatter_plot(
        self,
        df: pd.DataFrame,
        x_col: str,
        y_col: str,
        color_col: str,
        label_col: str,
        regression_line: bool = False,
        marker_size: int = 15,
    ) -> go.Figure:
        """Create a Plotly Scatter Plot figure.

        Args:
            df (pd.DataFrame): The dataframe containing the data.
            x_col (str): The column to use for the x-axis.
            y_col (str): The column to use for the y-axis.
            color_col (str): The column to use for the color scale.
            label_col (str): The column to use for point labels.
            regression_line (bool): Whether to include a regression line.
            marker_size (int): Size of the markers. Default is 15.

        Returns:
            go.Figure: A Plotly Figure object.
        """
        # Check if we need to show labels
        show_labels = label_col != "none" and len(df) < 1000

        # Helper to generate hover text for each point.
        def generate_hover_text(row):
            return "<br>".join([f"{col}: {row[col]}" for col in self.hover_columns])

        # Generate hover text for all points.
        hovertext = df.apply(generate_hover_text, axis=1)
        hovertemplate = "%{hovertext}<extra></extra>"
        hoverinfo = "none" if self.suppress_hover_display else None

        # Determine marker settings based on the type of the color column.
        if pd.api.types.is_numeric_dtype(df[color_col]):
            marker_color = df[color_col]
            colorbar = dict(title=color_col, thickness=10)
            # Single trace for numeric data.
            data = [
                go.Scattergl(
                    x=df[x_col],
                    y=df[y_col],
                    mode="markers",
                    text=df[label_col].astype(str) if show_labels else None,
                    textposition="top center",
                    hoverinfo=hoverinfo,
                    hovertext=hovertext,
                    hovertemplate=hovertemplate,
                    customdata=df[self.custom_data],
                    marker=dict(
                        size=marker_size,
                        color=marker_color,
                        colorscale=self.colorscale,
                        colorbar=colorbar,
                        opacity=0.8,
                        line=dict(color="rgba(0,0,0,0.25)", width=1),
                    ),
                )
            ]
            showlegend = False
        else:
            # For categorical data, create one trace per category so that a legend appears.
            categories = sorted(df[color_col].astype(str).unique())

            # Discrete colorscale using Plotly Express's qualitative palette.
            discrete_colors = px.colors.qualitative.Plotly
            data = []
            for i, cat in enumerate(categories):
                sub_df = df[df[color_col] == cat]
                sub_hovertext = hovertext.loc[sub_df.index]
                trace = go.Scattergl(
                    x=sub_df[x_col],
                    y=sub_df[y_col],
                    mode="markers",
                    text=sub_df[label_col] if show_labels else None,  # Add text if labels enabled
                    textposition="top center",  # Position labels above points
                    name=cat,
                    hoverinfo=hoverinfo,
                    hovertext=sub_hovertext,
                    hovertemplate=hovertemplate,
                    customdata=sub_df[self.custom_data],
                    marker=dict(
                        size=marker_size,
                        color=discrete_colors[i % len(discrete_colors)],
                        opacity=0.8,
                        line=dict(color="rgba(0,0,0,0.25)", width=1),
                    ),
                )
                data.append(trace)
            showlegend = True

        # Okay we have the data, now create the figure.
        # Note: We're going to add the prediction interval bands and regression line before
        #       the scatter plot data to ensure they appear below the scatter points.
        figure = go.Figure()
        if y_col == "prediction" or x_col == "prediction":
            figure = prediction_intervals(df, figure, x_col)

        # Add regression line if enabled.
        if regression_line:
            axis_min = min(df[x_col].min(), df[y_col].min())
            axis_max = max(df[x_col].max(), df[y_col].max())
            figure.add_shape(
                type="line",
                line=dict(width=4, color="rgba(128, 128, 128, 1.0)"),
                x0=axis_min,
                x1=axis_max,
                y0=axis_min,
                y1=axis_max,
            )

        # Now add the scatter plot data on last (on top)
        figure.add_traces(data)

        # Set up axes.
        if self.show_axes:
            xaxis = dict(
                title=dict(text=x_col, font=dict(size=16), standoff=15), tickformat=".2f", tickfont=dict(size=10)
            )
            yaxis = dict(
                title=dict(text=y_col, font=dict(size=16), standoff=25), tickformat=".2f", tickfont=dict(size=10)
            )
        else:
            xaxis = dict(visible=False)
            yaxis = dict(visible=False)

        # Update layout.
        figure.update_layout(
            margin={"t": 20, "b": 55, "r": 0, "l": 35, "pad": 0},
            xaxis=xaxis,
            yaxis=yaxis,
            showlegend=showlegend,
            dragmode="pan",
            modebar={"bgcolor": "rgba(0, 0, 0, 0)"},
            uirevision="constant",
        )

        return figure

    def register_internal_callbacks(self):
        """Register any internal callbacks for the plugin."""

        @callback(
            Output(f"{self.component_id}-graph", "figure", allow_duplicate=True),
            [
                Input(f"{self.component_id}-x-dropdown", "value"),
                Input(f"{self.component_id}-y-dropdown", "value"),
                Input(f"{self.component_id}-color-dropdown", "value"),
                Input(f"{self.component_id}-label-dropdown", "value"),
                Input(f"{self.component_id}-regression-line", "value"),
            ],
            prevent_initial_call=True,
        )
        def _update_scatter_plot(x_value, y_value, color_value, label_value, regression_line):
            """Update the Scatter Plot Graph based on the dropdown values."""

            # Check if the dataframe is not empty and the values are not None
            if not self.df.empty and x_value and y_value and color_value:
                # Update Plotly Scatter Plot with the label value
                figure = self.create_scatter_plot(self.df, x_value, y_value, color_value, label_value, regression_line)
                return figure

            raise PreventUpdate

        @callback(
            Output(f"{self.component_id}-overlay", "show"),
            Output(f"{self.component_id}-overlay", "bbox"),
            Output(f"{self.component_id}-overlay", "children"),
            Input(f"{self.component_id}-graph", "hoverData"),
        )
        def _scatter_overlay(hover_data):
            if hover_data is None:
                # Hide the overlay if no hover data
                return False, no_update, no_update

            # Extract bounding box from hoverData
            bbox = hover_data["points"][0]["bbox"]

            # Create an SVG with a circle at the center
            svg = """
            <svg xmlns="http://www.w3.org/2000/svg" width="100" height="100" style="overflow: visible;">
                <!-- Circle for the node -->
                <circle cx="50" cy="50" r="10" stroke="rgba(255, 255, 255, 1)" stroke-width="3" fill="none" />
            </svg>
            """

            # Encode the SVG as Base64
            encoded_svg = base64.b64encode(svg.encode("utf-8")).decode("utf-8")
            data_uri = f"data:image/svg+xml;base64,{encoded_svg}"

            # Use an img tag for the overlay
            svg_image = html.Img(src=data_uri, style={"width": "100px", "height": "100px"})

            # Get the center of the bounding box
            center_x = (bbox["x0"] + bbox["x1"]) / 2
            center_y = (bbox["y0"] + bbox["y1"]) / 2

            # The tooltip should be centered on the point (note: 'bottom' tooltip, so we adjust y position)
            adjusted_bbox = {
                "x0": center_x - 50,
                "x1": center_x + 50,
                "y0": center_y - 162,
                "y1": center_y - 62,
            }
            # Return the updated values for the overlay
            return True, adjusted_bbox, [svg_image]


if __name__ == "__main__":
    """Run the Unit Test for the Plugin."""
    from workbench.web_interface.components.plugin_unit_test import PluginUnitTest

    # Create a fake dataframe with 3 numeric columns and 2 categorical columns
    data = {
        "x": [1, 2, 3, 4, 5],
        "y": [2, 3, 4, 5, 6],
        "z": [3, 4, 5, 6, 7],
        "class": ["A", "C", "B", "B", "A"],
        "label": ["good", "bad", "okay", "good", "bad"],
    }
    df = pd.DataFrame(data)

    # Get a UQ regressor model
    # from workbench.api import Endpoint, DFStore
    # end = Endpoint("aqsol-uq")
    # df = end.auto_inference()
    # DFStore().upsert("/workbench/models/aqsol-uq/auto_inference", df)

    from workbench.api import DFStore

    df = DFStore().get("/workbench/models/aqsol-uq/auto_inference")

    # Run the Unit Test on the Plugin
    PluginUnitTest(
        ScatterPlot,
        input_data=df,
        theme="midnight_blue",
        x="solubility",
        y="prediction",
        color="residuals_abs",
        suppress_hover_display=True,
    ).run()



"""This Script creates the AQSol Ensemble Regression Artifacts in AWS/Workbench

Models:
    - aqsol-ensemble
Endpoints:
    - aqsol-ensemble
"""

import logging

from workbench.api import FeatureSet, Model, ModelType, Endpoint

log = logging.getLogger("workbench")


if __name__ == "__main__":

    # Recreate Flag in case you want to recreate the artifacts
    recreate = False

    # Create the AQSol Ensemble Regression Model
    if recreate or not Model("aqsol-ensemble").exists():

        # AQSol Features
        features = [
            "molwt",
            "mollogp",
            "molmr",
            "heavyatomcount",
            "numhacceptors",
            "numhdonors",
            "numheteroatoms",
            "numrotatablebonds",
            "numvalenceelectrons",
            "numaromaticrings",
            "numsaturatedrings",
            "numaliphaticrings",
            "ringcount",
            "tpsa",
            "labuteasa",
            "balabanj",
            "bertzct",
        ]

        # Transform FeatureSet into Quantile Regression Model
        feature_set = FeatureSet("aqsol_features")
        feature_set.to_model(
            name="aqsol-ensemble",
            model_type=ModelType.ENSEMBLE_REGRESSOR,
            target_column="solubility",
            feature_list=features,
            description="AQSol Ensemble Regression",
            tags=["aqsol", "ensemble"],
        )

    # Create the AQSol Quantile Regression Endpoint
    if recreate or not Endpoint("aqsol-ensemble").exists():
        m = Model("aqsol-ensemble")
        end = m.to_endpoint(tags=["aqsol", "ensemble"])

        # Run auto-inference on the AQSol Ensemble Regression Endpoint
        end.auto_inference(capture=True)



"""A proximity circle visualization component for comparing a query target to its neighbors"""

from dash import dcc
import plotly.graph_objects as go
import pandas as pd
import numpy as np

# Workbench Imports
from workbench.web_interface.components.plugin_interface import PluginInterface, PluginPage, PluginInputType
from workbench.utils.theme_manager import ThemeManager


class ProximityMiniGraph(PluginInterface):
    """Proximity Visualization Component for comparing a target and its nearest neighbors"""

    auto_load_page = PluginPage.NONE
    plugin_input_type = PluginInputType.DATAFRAME

    def __init__(self):
        """Initialize the ProximityMiniGraph plugin class"""
        self.component_id = None
        self.theme_manager = ThemeManager()
        self.colorscale = self.theme_manager.colorscale()
        super().__init__()

    def create_component(self, component_id: str) -> dcc.Graph:
        """Create a Proximity Visualization Component without any data."""
        self.component_id = component_id
        self.container = dcc.Graph(
            id=component_id,
            figure=self.display_text("Waiting for Data..."),
            config={"scrollZoom": False, "doubleClick": "reset", "displayModeBar": False},
        )
        self.properties = [(self.component_id, "figure")]
        self.signals = [(self.component_id, "clickData")]
        return self.container

    def update_properties(self, df: pd.DataFrame, **kwargs) -> list:
        """Create a Proximity Circle Figure for the query target and its neighbors."""
        # Basic validation
        if df is None or not isinstance(df, pd.DataFrame) or df.empty:
            return [self.display_text("No Data")]

        # Get column names
        id_column = df.columns[0]
        target_column = df.columns[-1]

        # Ensure required columns exist
        required_columns = [id_column, "neighbor_id", "distance", target_column]
        if not all(col in df.columns for col in required_columns):
            return [self.display_text("Missing required data columns")]

        # Get query ID and its neighbors
        query_id = df.iloc[0][id_column]
        query_target = df[df["neighbor_id"] == query_id].iloc[0][target_column]
        neighbors = df[df["neighbor_id"] != query_id].sort_values("distance").head(6)

        # Create figure
        fig = go.Figure()

        # Center coordinates and layout parameters
        center = (0, 0)
        radius = 0.5

        # Calculate coordinates for surrounding circles
        circle_coords = []
        for i in range(min(6, len(neighbors))):
            angle = (np.pi / 3) * i
            x = center[0] + radius * np.cos(angle)
            y = center[1] + radius * np.sin(angle)
            circle_coords.append((x, y))

        # Prepare data for visualization
        x_coords, y_coords = [], []
        lines_data = []
        node_data = []

        # Set up center node
        x_coords.append(center[0])
        y_coords.append(center[1])
        node_data.append(
            {
                "id": query_id,
                "color": 0,
                "size": 40,
                "hover_text": f"ID: {query_id}<br>Value: {query_target:.2f}",
                "text": f"{query_target:.2f}",
                "type": "query",
                "distance": 0,
                "target": query_target,
            }
        )

        # Set up neighbor nodes and connecting lines
        for i, (x, y) in enumerate(circle_coords):
            neighbor = neighbors.iloc[i]
            neighbor_id = neighbor["neighbor_id"]
            distance = neighbor["distance"]
            target_val = neighbor[target_column]
            delta = target_val - query_target

            # Add node data
            x_coords.append(x)
            y_coords.append(y)
            node_data.append(
                {
                    "id": neighbor_id,
                    "color": delta,
                    "size": 40,
                    "hover_text": f"ID: {neighbor_id}<br>Distance: {distance:.2f}<br>Value: {target_val:.2f}",
                    "text": f"{target_val:.2f}",
                    "type": "neighbor",
                    "distance": distance,
                    "target": target_val,
                }
            )

            # Add line data
            line_width = max(4 - distance, 0) * 5 + 5
            lines_data.append({"x": [center[0], x], "y": [center[1], y], "width": line_width})

        # Add connecting lines to figure
        for line in lines_data:
            fig.add_trace(
                go.Scatter(
                    x=line["x"],
                    y=line["y"],
                    mode="lines",
                    line=dict(color="rgba(150, 150, 150, 0.75)", width=line["width"]),
                    hoverinfo="none",
                    showlegend=False,
                )
            )

        # Extract data for scatter plot
        colors = [node["color"] for node in node_data]
        sizes = [node["size"] for node in node_data]
        hover_texts = [node["hover_text"] for node in node_data]
        texts = [node["text"] for node in node_data]

        # Create custom data for interactivity
        custom_data = []
        for node in node_data:
            custom_data.append(
                {"type": node["type"], "id": node["id"], "distance": node["distance"], "target": node["target"]}
            )

        # Add nodes to figure
        scatter = go.Scatter(
            x=x_coords,
            y=y_coords,
            mode="markers+text",
            marker=dict(
                color=colors, colorscale=self.colorscale, size=sizes, line=dict(color="black", width=1), opacity=1.0
            ),
            text=texts,
            hovertext=hover_texts,
            hoverinfo="text",
            showlegend=False,
        )
        scatter.customdata = custom_data
        fig.add_trace(scatter)

        # Configure layout
        fig.update_layout(
            margin=dict(l=0, r=0, t=0, b=0),
            height=200,
            width=200,
            showlegend=False,
            plot_bgcolor=self.theme_manager.background(),
            hoverlabel=dict(bgcolor="white", font_size=12, font_family="Arial"),
            hovermode="closest",
            xaxis=dict(range=[-1, 1], visible=False),
            yaxis=dict(range=[-1, 1], visible=False, scaleanchor="x", scaleratio=1),
        )

        return [fig]

    def register_internal_callbacks(self):
        """Register internal callbacks for the plugin."""
        pass  # We'll implement this if needed


if __name__ == "__main__":
    """Run the Unit Test for the Plugin."""
    from workbench.web_interface.components.plugin_unit_test import PluginUnitTest

    # Create sample data for testing
    test_data = pd.DataFrame(
        {
            "id": ["1", "1", "1", "1", "1", "1", "1"],
            "neighbor_id": ["1", "2", "3", "4", "5", "6", "7"],
            "distance": [0.0, 0.0, 1.299653, 1.377522, 1.442327, 2.636339, 5.636339],
            "activity": [4.2, 8.0, 3.499257, 4.368489, 4.34046, 8.0, 2.0],
        }
    )

    # Run the Unit Test on the Plugin
    PluginUnitTest(ProximityMiniGraph, input_data=test_data, theme="light").run()



"""An Outlier Plot Component"""

import plotly.graph_objs
from dash import dcc
import pandas as pd
import numpy as np
import plotly.express as px
import logging

# Workbench Imports
from workbench.algorithms.dataframe.aggregation import aggregate
from workbench.algorithms.dataframe.projection_2d import Projection2D


# Workbench Logger
log = logging.getLogger("workbench")


# For colormaps see (https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express)
def create_figure(df: pd.DataFrame, title: str = "Outlier Groups") -> plotly.graph_objs.Figure:
    """Create a Outlier Plot
    Args:
        df (pd.DataFrame): The dataframe containing the data.
        title (str): The title for the plot
    Returns:
        plotly.graph_objs.Figure: A Figure object containing the generated plots.
    """

    # The dataframe is required to have the following columns:
    #   outlier_group: The outlier group descriptor (e.g. "mol_log_p_high")
    if "outlier_group" not in df.columns:
        raise ValueError("Outlier Plot requires column 'outlier_group' to be in the dataframe.")

    # Now process the dataframe
    # - aggregate the dataframe by outlier_group
    # - compute the mean of each group
    # - run dimensionality reduction on the mean values
    if "x" not in df.columns:
        log.info("Outlier Plot: No coordinates in dataframe, running aggregation and dimensionality reduction...")
        agg_df = aggregate(df, group_column="outlier_group")
        coord_df = Projection2D().fit_transform(agg_df)
    else:
        log.info("Outlier Plot: Coordinates found...")
        coord_df = df

    # Scale the group_count column so that it's between 5 and 100
    min_size = 5
    max_size = 100
    coord_df["group_count"] = np.interp(
        coord_df["group_count"],
        (coord_df["group_count"].min(), coord_df["group_count"].max()),
        (min_size, max_size),
    )

    # Sample column needs to be first
    coord_df["is_sample"] = coord_df["outlier_group"] == "sample"
    coord_df.sort_values("is_sample", ascending=False, inplace=True)
    coord_df.drop("is_sample", axis=1, inplace=True)

    # Create the Outlier Plot
    color_map = px.colors.qualitative.Plotly
    fig = px.scatter(
        coord_df,
        x="x",
        y="y",
        opacity=0.75,
        color="outlier_group",
        size="group_count",
        size_max=100,
        title=title,
        color_discrete_sequence=color_map,
    )

    # Make sure the 'sample' group is always grey
    for trace in fig.data:
        if "sample" in trace.name:
            trace.marker.color = "darkgrey"

    # Various Plotly Layout Updates
    fig.update_layout(title_y=0.97, title_x=0.3, title_xanchor="center", title_yanchor="top")
    fig.update_traces(
        marker=dict(line=dict(width=2, color="black")),
        selector=dict(mode="markers"),
    )
    fig.update_layout(xaxis_visible=False, xaxis_showticklabels=False)
    fig.update_layout(yaxis_visible=False, yaxis_showticklabels=False)
    fig.update_layout(margin=dict(l=10, r=10, t=30, b=10), height=400)
    return fig


def create(component_id: str, df: pd.DataFrame, title: str = "Outlier Groups") -> dcc.Graph:
    """Create a Graph Component for outlier plots.

    Args:
        component_id (str): The ID of the UI component
        df (pd.DataFrame): A dataframe of data
        title (str): The title for the plot
    Returns:
        dcc.Graph: A Dash Graph Component representing the vertical distribution plots.
    """

    # Generate a figure and wrap it in a Dash Graph Component
    return dcc.Graph(id=component_id, figure=create_figure(df, title=title))



"""Pipelines:  A Workbench Web Interface to view and interact with Pipelines"""

from dash import register_page

# Local Imports
from .layout import pipelines_layout
from . import callbacks

# Workbench Imports
from workbench.web_interface.components.plugins import pipeline_details, ag_table
from workbench.web_interface.components.plugin_interface import PluginPage
from workbench.web_interface.page_views.pipelines_page_view import PipelinesPageView
from workbench.utils.plugin_manager import PluginManager

# Register this page with Dash
register_page(
    __name__,
    path="/pipelines",
    name="Workbench - Pipelines",
)

# Create a table to display the pipelines
pipeline_table = ag_table.AGTable()
table_component = pipeline_table.create_component("pipelines_table")

# Create a Markdown component to display the pipeline details
pipeline_details = pipeline_details.PipelineDetails()
details_component = pipeline_details.create_component("pipeline_details")

# Capture our components in a dictionary to send off to the layout
components = {"pipelines_table": table_component, "pipeline_details": details_component}

# Load any web components plugins of type 'pipeline'
pm = PluginManager()
plugins = pm.get_list_of_web_plugins(plugin_page=PluginPage.PIPELINE)

# Add the plugins to the components dictionary
for plugin in plugins:
    component_id = plugin.generate_component_id()
    components[component_id] = plugin.create_component(component_id)

# Set up our layout (Dash looks for a var called layout)
layout = pipelines_layout(**components)

# Grab a view that gives us a summary of the Pipelines in Workbench
pipelines_view = PipelinesPageView()

# Callback for anything we want to happen on page load
callbacks.on_page_load()

# Setup our callbacks/connections
callbacks.pipeline_table_refresh(pipelines_view, pipeline_table)

# We're going to add the details component to the plugins list
plugins.append(pipeline_details)

# For all the plugins we have we'll call their update_properties method
if plugins:
    callbacks.setup_plugin_callbacks(plugins)



from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp
import time
from botocore.exceptions import ClientError
import logging

# Initialize logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Initialize a session using your AWS credentials
aws_account_clamp = AWSAccountClamp()
session = aws_account_clamp.boto3_session
glue_client = session.client("glue")

# Specify your source and destination databases
source_database = "sageworks"
destination_database = "workbench"


def ensure_database_exists(db_name):
    try:
        glue_client.get_database(Name=db_name)
        logger.info(f"Database {db_name} already exists.")
    except glue_client.exceptions.EntityNotFoundException:
        glue_client.create_database(DatabaseInput={"Name": db_name})
        logger.info(f"Database {db_name} created successfully.")


def sanitize_storage_descriptor(storage_descriptor):
    unsupported_fields = ["ColumnsStatistics", "SortColumns"]  # Add more if necessary
    for field in unsupported_fields:
        storage_descriptor.pop(field, None)
    return storage_descriptor


def create_table_with_retry(dest_db, table_input, retries=3, delay=5):
    for attempt in range(retries):
        try:
            glue_client.create_table(DatabaseName=dest_db, TableInput=table_input)
            return True
        except ClientError as e:
            if attempt < retries - 1:
                logger.warning(f"Retrying table {table_input['Name']}... ({attempt+1}/{retries})")
                time.sleep(delay)
            else:
                raise e


def flip_parameters(parameters: dict) -> dict:
    return {key.replace("sageworks", "workbench"): value for key, value in parameters.items()}


def copy_tables(src_db, dest_db):
    # Ensure destination database exists
    ensure_database_exists(dest_db)

    # Get the list of tables from the source database
    paginator = glue_client.get_paginator("get_tables")
    for page in paginator.paginate(DatabaseName=src_db):
        for table in page["TableList"]:
            # Flip 'sageworks' to 'workbench' in parameters
            updated_parameters = flip_parameters(table.get("Parameters", {}))

            # Construct the TableInput
            table_input = {
                "Name": table["Name"],
                "Description": table.get("Description", ""),
                "Retention": table.get("Retention", 0),
                "StorageDescriptor": sanitize_storage_descriptor(table["StorageDescriptor"]),
                "PartitionKeys": table.get("PartitionKeys", []),
                "TableType": table.get("TableType", ""),
                "Parameters": updated_parameters,
            }

            # Add view-specific fields
            if "ViewOriginalText" in table:
                table_input["ViewOriginalText"] = table["ViewOriginalText"]
            if "ViewExpandedText" in table:
                table_input["ViewExpandedText"] = table["ViewExpandedText"]

            try:
                create_table_with_retry(dest_db, table_input)
                logger.info(f"Table {table['Name']} copied successfully.")
            except glue_client.exceptions.AlreadyExistsException:
                logger.warning(f"Table {table['Name']} already exists in the destination database.")
            except Exception as e:
                logger.error(f"Error copying table {table['Name']}: {e}")


# Run the function to start copying tables
copy_tables(source_database, destination_database)



from fastapi import FastAPI, Request, Response
from contextlib import asynccontextmanager
import os
import sys
import json
import importlib.util
import logging
import subprocess
import site

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Global variables
model = None
inference_module = None


def get_inference_script(model_dir: str) -> str:
    """Retrieve the inference script name

    Args:
        model_dir (str): The directory containing the model artifacts

    Returns:
        str: The name of the inference script
    """

    # Get the path to the inference-metadata.json file
    inference_meta_path = os.path.join(model_dir, "inference-metadata.json")
    with open(inference_meta_path, "r") as f:
        config = json.load(f)
        return config["inference_script"]


def install_requirements(requirements_path):
    """Install Python dependencies from requirements file.
    Uses a persistent cache to speed up container cold starts.
    Note: Inference containers don't have root access, so we
          use the --user flag and add the user package path manually.
    """
    if os.path.exists(requirements_path):
        logger.info(f"Installing dependencies from {requirements_path}...")

        # Define a persistent cache location
        pip_cache_dir = "/opt/ml/model/.cache/pip"
        os.environ["PIP_CACHE_DIR"] = pip_cache_dir

        try:
            subprocess.check_call(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--cache-dir",
                    pip_cache_dir,  # Enable caching
                    "--disable-pip-version-check",
                    "--no-warn-script-location",
                    "--user",
                    "-r",
                    requirements_path,
                ]
            )
            # Ensure Python can find user-installed packages
            sys.path.append(site.getusersitepackages())
            logger.info("Requirements installed successfully.")
        except subprocess.CalledProcessError as e:
            logger.error(f"Error installing requirements: {e}")
            sys.exit(1)
    else:
        logger.info(f"No requirements file found at {requirements_path}")


@asynccontextmanager
async def lifespan(app: FastAPI):
    """Handle model loading on startup and cleanup on shutdown."""
    global model, inference_module

    # Note: SageMaker will put model.tar.gz in /opt/ml/model
    #       which includes the model artifacts and inference code
    model_dir = os.environ.get("SM_MODEL_DIR", "/opt/ml/model")
    inference_script = get_inference_script(model_dir)

    # List directory contents for debugging
    logger.info(f"Contents of {model_dir}: {os.listdir(model_dir)}")

    try:
        # Load the inference script from source_dir
        inference_script_path = os.path.join(model_dir, inference_script)
        if not os.path.exists(inference_script_path):
            raise FileNotFoundError(f"Inference script not found: {inference_script_path}")

        # Install requirements if present
        install_requirements(os.path.join(model_dir, "requirements.txt"))

        # Ensure the model directory is in the Python path
        sys.path.insert(0, model_dir)

        # Import the inference module
        logger.info(f"Importing inference module from {inference_script_path}")
        spec = importlib.util.spec_from_file_location("inference_module", inference_script_path)
        inference_module = importlib.util.module_from_spec(spec)
        sys.modules["inference_module"] = inference_module
        spec.loader.exec_module(inference_module)

        # Check if model_fn is defined
        if not hasattr(inference_module, "model_fn"):
            raise ImportError(f"Inference module {inference_script_path} does not define model_fn")

        # Load the model using model_fn
        logger.info("Calling model_fn to load the model")
        model = inference_module.model_fn(model_dir)
        logger.info(f"Model loaded successfully: {type(model)}")

    except Exception as e:
        logger.error(f"Error initializing model: {e}", exc_info=True)
        raise

    yield

    logger.info("Shutting down model server")


app = FastAPI(lifespan=lifespan)


@app.get("/ping")
def ping():
    """Health check endpoint for SageMaker."""
    # Check if the inference module is loaded
    return Response(status_code=200 if inference_module else 500)


@app.post("/invocations")
async def invoke(request: Request):
    """Inference endpoint for SageMaker."""
    content_type = request.headers.get("Content-Type", "")
    accept_type = request.headers.get("Accept", "")

    try:
        body = await request.body()
        data = inference_module.input_fn(body, content_type)
        result = inference_module.predict_fn(data, model)
        output_data, output_content_type = inference_module.output_fn(result, accept_type)
        return Response(content=output_data, media_type=output_content_type)
    except Exception as e:
        logger.error(f"Error during inference: {e}", exc_info=True)
        return Response(content=json.dumps({"error": str(e)}), status_code=500, media_type="application/json")



"""Tests for the Data to Data (light) Transforms"""

import pytest

# Local imports
from workbench.core.transforms.data_to_features.light.data_to_features_light import DataToFeaturesLight
from workbench.api.data_source import DataSource


# Simple test of the DataToFeaturesLight functionality
@pytest.mark.long
def transform_test():
    """Tests for the Data to Features (light) Transforms"""

    # Create the class with inputs and outputs and invoke the transform
    input_name = "test_data"
    output_name = "test_features_temp"
    data_to_features = DataToFeaturesLight(input_name, output_name)
    data_to_features.set_output_tags(["temp", "test", "small"])
    data_to_features.transform(id_column="id", event_time_column="date")


# Testing the DataSource API to_features() method
@pytest.mark.long
def to_features_test():

    ds = DataSource("wine_data")
    ds.to_features("wine_features_temp", tags=["temp", "wine", "classification"])


if __name__ == "__main__":
    transform_test()
    to_features_test()



from pprint import pprint
from workbench.api.feature_set import FeatureSet
from workbench.api.model import Model, ModelType
from workbench.utils.model_utils import get_custom_script_path

# Grab a FeatureSet
my_features = FeatureSet("aqsol_features")

# Grab our AQSol Regression Model
model = Model("aqsol-regression")

# Get the custom script path for the UQ model
script_path = get_custom_script_path("uq_models", "ngboost.template")

# Get Feature and Target Columns from the existing given Model
features = model.features()
target = model.target()

# Create the NGBoost Model from our FeatureSet
fs = FeatureSet(model.get_input())
my_model = fs.to_model(
    name="aqsol-ngboost-reg",
    model_type=ModelType.UQ_REGRESSOR,
    feature_list=features,
    target_column=target,
    description=f"NGBoost Model for AQSol",
    tags=["aqsol", "ngboost", "regression"],
    custom_script=script_path,
)
# Print the details of the created model
pprint(my_model.details())
my_model = Model("aqsol-ngboost-reg")

# Deploy an Endpoint for the Model
endpoint = my_model.to_endpoint(tags=["aqsol", "ngboost"])

# Run auto-inference on the Endpoint
endpoint.auto_inference(capture=True)



"""PandasToFeatures: Class to publish a Pandas DataFrame into a FeatureSet"""

import pandas as pd
import time
import re
from sagemaker.feature_store.feature_group import FeatureGroup, IngestionError
from sagemaker.feature_store.inputs import TableFormatEnum

# Local imports
from workbench.utils.datetime_utils import datetime_to_iso8601
from workbench.core.transforms.transform import Transform, TransformInput, TransformOutput
from workbench.core.artifacts.artifact import Artifact
from workbench.core.artifacts.feature_set_core import FeatureSetCore


class PandasToFeatures(Transform):
    """PandasToFeatures: Class to publish a Pandas DataFrame into a FeatureSet

    Common Usage:
        ```python
        to_features = PandasToFeatures(output_name)
        to_features.set_output_tags(["my", "awesome", "data"])
        to_features.set_input(df, id_column="my_id")
        to_features.transform()
        ```
    """

    def __init__(self, output_name: str):
        """PandasToFeatures Initialization

        Args:
            output_name (str): The Name of the FeatureSet to create
        """

        # Make sure the output_name is a valid name
        Artifact.is_name_valid(output_name)

        # Call superclass init
        super().__init__("DataFrame", output_name)

        # Set up all my instance attributes
        self.input_type = TransformInput.PANDAS_DF
        self.output_type = TransformOutput.FEATURE_SET
        self.id_column = None
        self.event_time_column = None
        self.one_hot_columns = []
        self.categorical_dtypes = {}  # Used for streaming/chunking
        self.output_df = None
        self.table_format = TableFormatEnum.ICEBERG
        self.incoming_hold_out_ids = None

        # These will be set in the transform method
        self.output_feature_group = None
        self.output_feature_set = None
        self.expected_rows = 0

    def set_input(self, input_df: pd.DataFrame, id_column=None, event_time_column=None, one_hot_columns=None):
        """Set the Input DataFrame for this Transform

        Args:
            input_df (pd.DataFrame): The input DataFrame.
            id_column (str, optional): The ID column (if not specified, an 'auto_id' will be generated).
            event_time_column (str, optional): The name of the event time column (default: None).
            one_hot_columns (list, optional): The list of columns to one-hot encode (default: None).
        """
        self.id_column = id_column
        self.event_time_column = event_time_column
        self.output_df = input_df.copy()
        self.one_hot_columns = one_hot_columns or []

        # Now Prepare the DataFrame for its journey into an AWS FeatureGroup
        self.prep_dataframe()

    def delete_existing(self):
        # Delete the existing FeatureSet if it exists
        self.log.info(f"Deleting the {self.output_name} FeatureSet...")
        FeatureSetCore.managed_delete(self.output_name)
        time.sleep(1)

    def _ensure_id_column(self):
        """Internal: AWS Feature Store requires an Id field"""
        if self.id_column is None:
            self.log.warning("Generating an 'auto_id' column, we highly recommended setting the 'id_column'")
            self.output_df["auto_id"] = self.output_df.index
            self.id_column = "auto_id"
            return
        if self.id_column not in self.output_df.columns:
            error_msg = f"Id column {self.id_column} not found in the DataFrame"
            self.log.critical(error_msg)
            raise ValueError(error_msg)

    def _ensure_event_time(self):
        """Internal: AWS Feature Store requires an event_time field for all data stored"""
        if self.event_time_column is None or self.event_time_column not in self.output_df.columns:
            self.log.info("Generating an event_time column before FeatureSet Creation...")
            self.event_time_column = "event_time"
            self.output_df[self.event_time_column] = pd.Timestamp("now", tz="UTC")

        # The event_time_column is defined, so we need to make sure it's in ISO-8601 string format
        # Note: AWS Feature Store only a particular ISO-8601 format not ALL ISO-8601 formats
        time_column = self.output_df[self.event_time_column]

        # Check if the event_time_column is of type object or string convert it to DateTime
        if time_column.dtypes == "object" or time_column.dtypes.name == "string":
            self.log.info(f"Converting {self.event_time_column} to DateTime...")
            time_column = pd.to_datetime(time_column)

        # Let's make sure it the right type for Feature Store
        if pd.api.types.is_datetime64_any_dtype(time_column):
            self.log.info(f"Converting {self.event_time_column} to ISOFormat Date String before FeatureSet Creation...")

            # Convert the datetime DType to ISO-8601 string
            # TableFormat=ICEBERG does not support alternate formats for event_time field, it only supports String type.
            time_column = time_column.map(datetime_to_iso8601)
            self.output_df[self.event_time_column] = time_column.astype("string")

    def _convert_objs_to_string(self):
        """Internal: AWS Feature Store doesn't know how to store object dtypes, so convert to String"""
        for col in self.output_df:
            if pd.api.types.is_object_dtype(self.output_df[col].dtype):
                self.output_df[col] = self.output_df[col].astype(pd.StringDtype())

    def process_column_name(self, column: str, shorten: bool = False) -> str:
        """Call various methods to make sure the column is ready for Feature Store
        Args:
            column (str): The column name to process
            shorten (bool): Should we shorten the column name? (default: False)
        """
        self.log.debug(f"Processing column {column}...")

        # Make sure the column name is valid
        column = self.sanitize_column_name(column)

        # Make sure the column name isn't too long
        if shorten:
            column = self.shorten_column_name(column)

        return column

    def shorten_column_name(self, name, max_length=20):
        if len(name) <= max_length:
            return name

        # Start building the new name from the end
        parts = name.split("_")[::-1]
        new_name = ""
        for part in parts:
            if len(new_name) + len(part) + 1 <= max_length:  # +1 for the underscore
                new_name = f"{part}_{new_name}" if new_name else part
            else:
                break

        # If new_name is empty, just use the last part of the original name
        if not new_name:
            new_name = parts[0]

        self.log.info(f"Shortening {name} to {new_name}")
        return new_name

    def sanitize_column_name(self, name):
        # Remove all invalid characters
        sanitized = re.sub("[^a-zA-Z0-9-_]", "_", name)
        sanitized = re.sub("_+", "_", sanitized)
        sanitized = sanitized.strip("_")

        # Log the change if the name was altered
        if sanitized != name:
            self.log.info(f"Sanitizing {name} to {sanitized}")

        return sanitized

    def one_hot_encode(self, df, one_hot_columns: list) -> pd.DataFrame:
        """One Hot Encoding for Categorical Columns with additional column name management

        Args:
            df (pd.DataFrame): The DataFrame to process
            one_hot_columns (list): The list of columns to one-hot encode

        Returns:
            pd.DataFrame: The DataFrame with one-hot encoded columns
        """

        # Grab the current list of columns
        current_columns = list(df.columns)

        # Now convert the list of columns into Categorical and then One-Hot Encode
        self.convert_columns_to_categorical(one_hot_columns)
        self.log.important(f"One-Hot encoding columns: {one_hot_columns}")
        df = pd.get_dummies(df, columns=one_hot_columns)

        # Compute the new columns generated by get_dummies
        new_columns = list(set(df.columns) - set(current_columns))
        self.log.important(f"New columns generated: {new_columns}")

        # Convert new columns to int32
        df[new_columns] = df[new_columns].astype("int32")

        # For the new columns we're going to shorten the names
        renamed_columns = {col: self.process_column_name(col) for col in new_columns}

        # Rename the columns in the DataFrame
        df.rename(columns=renamed_columns, inplace=True)

        return df

    # Helper Methods
    def convert_columns_to_categorical(self, columns: list):
        """Convert column to Categorical type"""
        for feature in columns:
            if feature not in [self.event_time_column, self.id_column]:
                unique_values = self.output_df[feature].nunique()
                if 1 < unique_values < 10:
                    self.log.important(f"Converting column {feature} to categorical (unique {unique_values})")
                    self.output_df[feature] = self.output_df[feature].astype("category")
                else:
                    self.log.warning(f"Column {feature} too many unique values {unique_values} skipping...")

    def manual_categorical_converter(self):
        """Used for Streaming: Convert object and string types to Categorical

        Note:
            This method is used for streaming/chunking. You can set the
            categorical_dtypes attribute to a dictionary of column names and
            their respective categorical types.
        """
        for column, cat_d_type in self.categorical_dtypes.items():
            self.output_df[column] = self.output_df[column].astype(cat_d_type)

    @staticmethod
    def convert_column_types(df: pd.DataFrame) -> pd.DataFrame:
        """Convert the types of the DataFrame to the correct types for the Feature Store"""
        for column in list(df.select_dtypes(include="bool").columns):
            df[column] = df[column].astype("int32")
        for column in list(df.select_dtypes(include="category").columns):
            df[column] = df[column].astype("str")

        # Select all columns that are of datetime dtype and convert them to ISO-8601 strings
        for column in [col for col in df.columns if pd.api.types.is_datetime64_any_dtype(df[col])]:
            df[column] = df[column].map(datetime_to_iso8601).astype("string")

        """FIXME Not sure we need these conversions
        for column in list(df.select_dtypes(include="object").columns):
            df[column] = df[column].astype("string")
        for column in list(df.select_dtypes(include=[pd.Int64Dtype]).columns):
            df[column] = df[column].astype("int64")
        for column in list(df.select_dtypes(include=[pd.Float64Dtype]).columns):
            df[column] = df[column].astype("float64")
        """
        return df

    def prep_dataframe(self):
        """Prep the DataFrame for Feature Store Creation"""
        self.log.info("Prep the output_df (cat_convert, convert types, and lowercase columns)...")

        # Remove any columns generated from AWS
        aws_cols = ["write_time", "api_invocation_time", "is_deleted", "event_time"]
        self.output_df = self.output_df.drop(columns=aws_cols, errors="ignore")

        # If one-hot columns are provided then one-hot encode them
        if self.one_hot_columns:
            self.output_df = self.one_hot_encode(self.output_df, self.one_hot_columns)

        # Convert columns names to lowercase, Athena will not work with uppercase column names
        if str(self.output_df.columns) != str(self.output_df.columns.str.lower()):
            for c in self.output_df.columns:
                if c != c.lower():
                    self.log.important(f"Column name {c} converted to lowercase: {c.lower()}")
            self.output_df.columns = self.output_df.columns.str.lower()

        # Check for duplicate column names in the dataframe
        if len(self.output_df.columns) != len(set(self.output_df.columns)):
            self.log.critical("Duplicate column names detected in the DataFrame")
            duplicates = self.output_df.columns[self.output_df.columns.duplicated()].tolist()
            self.log.critical(f"Duplicated columns: {duplicates}")
            raise ValueError("Duplicate column names detected in the DataFrame")

        # Make sure we have the required id and event_time columns
        self._ensure_id_column()
        self._ensure_event_time()

        # Check for a training column (Workbench uses dynamic training columns)
        if "training" in self.output_df.columns:
            self.log.important(
                """Training column detected: Since FeatureSets are read-only, Workbench creates a training view
                that can be dynamically changed. We'll use this training column to create a training view."""
            )
            self.incoming_hold_out_ids = self.output_df[~self.output_df["training"]][self.id_column].tolist()
            self.output_df = self.output_df.drop(columns=["training"])

        # We need to convert some of our column types to the correct types
        # Feature Store only supports these data types:
        # - Integral
        # - Fractional
        # - String (timestamp/datetime types need to be converted to string)
        self.output_df = self.convert_column_types(self.output_df)

    def create_feature_group(self):
        """Create a Feature Group, load our Feature Definitions, and wait for it to be ready"""

        # Create a Feature Group and load our Feature Definitions
        my_feature_group = FeatureGroup(name=self.output_name, sagemaker_session=self.sm_session)
        my_feature_group.load_feature_definitions(data_frame=self.output_df)

        # Create the Output S3 Storage Path for this Feature Set
        s3_storage_path = f"{self.feature_sets_s3_path}/{self.output_name}"

        # Get the metadata/tags to push into AWS
        aws_tags = self.get_aws_tags()

        # Create the Feature Group
        my_feature_group.create(
            s3_uri=s3_storage_path,
            record_identifier_name=self.id_column,
            event_time_feature_name=self.event_time_column,
            role_arn=self.workbench_role_arn,
            enable_online_store=True,
            table_format=self.table_format,
            tags=aws_tags,
        )

        # Ensure/wait for the feature group to be created
        self.ensure_feature_group_created(my_feature_group)
        return my_feature_group

    def pre_transform(self, **kwargs):
        """Pre-Transform: Delete any existing FeatureSet and Create the Feature Group"""
        self.delete_existing()
        self.output_feature_group = self.create_feature_group()

    def transform_impl(self):
        """Transform Implementation: Ingest the data into the Feature Group"""

        # Now we actually push the data into the Feature Group (called ingestion)
        self.log.important(f"Ingesting rows into Feature Group {self.output_name}...")
        ingest_manager = self.output_feature_group.ingest(self.output_df, max_workers=8, max_processes=4, wait=False)
        try:
            ingest_manager.wait()
        except IngestionError as exc:
            self.log.warning(f"Some rows had an ingesting error: {exc}")

        # Report on any rows that failed to ingest
        if ingest_manager.failed_rows:
            self.log.warning(f"Number of Failed Rows: {len(ingest_manager.failed_rows)}")

            # Log failed row details
            failed_data = self.output_df.iloc[ingest_manager.failed_rows]
            for idx, row in failed_data.iterrows():
                self.log.warning(f"Failed Row {idx}: {row.to_dict()}")

        # Keep track of the number of rows we expect to be ingested
        self.expected_rows += len(self.output_df) - len(ingest_manager.failed_rows)
        self.log.info(f"Added rows: {len(self.output_df)}")
        self.log.info(f"Failed rows: {len(ingest_manager.failed_rows)}")
        self.log.info(f"Total rows ingested: {self.expected_rows}")

        # We often need to wait a bit for AWS to fully register the new Feature Group
        self.log.important(f"Waiting for AWS to register the new Feature Group {self.output_name}...")
        time.sleep(30)

    def post_transform(self, **kwargs):
        """Post-Transform: Populating Offline Storage and onboard()"""
        self.log.info("Post-Transform: Populating Offline Storage and onboard()...")

        # Feature Group Ingestion takes a while, so we need to wait for it to finish
        self.output_feature_set = FeatureSetCore(self.output_name)
        self.log.important("Waiting for AWS Feature Group Offline storage to be ready...")
        self.log.important("This will often take 10-20 minutes...go have coffee or lunch :)")
        self.output_feature_set.set_status("initializing")
        self.wait_for_rows(self.expected_rows)

        # Call the FeatureSet onboard method to compute a bunch of EDA stuff
        self.output_feature_set.onboard()

        # Set Hold Out Ids (if we got them during creation)
        if self.incoming_hold_out_ids:
            self.output_feature_set.set_training_holdouts(self.id_column, self.incoming_hold_out_ids)

    def ensure_feature_group_created(self, feature_group):
        status = feature_group.describe().get("FeatureGroupStatus")
        while status == "Creating":
            self.log.debug("FeatureSet being Created…")
            time.sleep(5)
            status = feature_group.describe().get("FeatureGroupStatus")

        if status == "Created":
            self.log.info(f"FeatureSet {feature_group.name} successfully created")
        else:
            # Get the detailed failure reason
            description = feature_group.describe()
            failure_reason = description.get("FailureReason", "No failure reason provided")
            self.log.critical(f"FeatureSet {feature_group.name} creation failed with status: {status}")
            self.log.critical(f"Failure reason: {failure_reason}")

    def wait_for_rows(self, expected_rows: int):
        """Wait for AWS Feature Group to fully populate the Offline Storage"""
        rows = self.output_feature_set.num_rows()

        # Wait for the rows to be populated
        self.log.info(f"Waiting for AWS Feature Group {self.output_name} Offline Storage...")
        max_retry = 20
        num_retry = 0
        sleep_time = 30
        while rows < expected_rows and num_retry < max_retry:
            num_retry += 1
            time.sleep(sleep_time)
            rows = self.output_feature_set.num_rows()
            self.log.info(f"Checking Offline Storage {self.output_name}: {rows}/{expected_rows} rows")
        if rows == expected_rows:
            self.log.important(f"Success: Reached Expected Rows ({rows} rows)...")
        else:
            msg = f"Did not reach expected rows ({rows}/{expected_rows})...(probably AWS lag)"
            self.log.warning(msg)
            self.log.monitor(msg)


if __name__ == "__main__":
    """Exercise the PandasToFeatures Class"""
    from workbench.api.data_source import DataSource

    # Setup Pandas output options
    pd.set_option("display.max_colwidth", 15)
    pd.set_option("display.max_columns", 15)
    pd.set_option("display.width", 1000)

    # Temp
    fs = FeatureSetCore("test_features")

    # Grab the test_data DataSource
    ds = DataSource("test_data")
    data_df = ds.sample()

    # Test setting a training column
    data_df["training"] = False
    data_df.loc[0:80, "training"] = True

    # Create my DF to Feature Set Transform (with one-hot encoding)
    df_to_features = PandasToFeatures("test_features")
    df_to_features.set_input(data_df, id_column="id", one_hot_columns=["food"])
    df_to_features.set_output_tags(["test", "small"])
    df_to_features.transform()

    # Test non-compliant output Name
    df_to_features = PandasToFeatures("test_features-123")

    #
    # Individual Tests
    #
    """
    # Test converting columns to categorical
    df_to_features = PandasToFeatures("test_features")
    df_to_features.set_input(data_df, id_column="id")
    df_to_features.convert_columns_to_categorical(["food", "likes_dogs"])

    # Test the one-hot encoding
    df_to_features.set_input(data_df, id_column="id", one_hot_columns=["food", "likes_dogs"])
    one_hot_df = df_to_features.one_hot_encode(data_df, ["food", "likes_dogs"])
    print(one_hot_df)
    """



"""AWSSecretsManager: Manages Workbench secrets in AWS Secrets Manager."""

from typing import Union
import logging
import json
from botocore.exceptions import ClientError

# Workbench Imports
from workbench.core.cloud_platform.aws.aws_session import AWSSession


class AWSSecretsManager:
    """AWSSecretsManager: Manages Workbench secrets in AWS Secrets Manager.

    Common Usage:
        ```python
        secrets = AWSSecretsManager()

        # List Secrets
        secrets.list()

        ['/workbench/api_key',
         '/workbench/database_credentials',
         '/workbench/oauth_token',
         '/workbench/pipelines/service_account']

        # Add Secret
        secrets.upsert("api_key", "my-secret-key")
        value = secrets.get("api_key")

        # Add structured data
        credentials = {"username": "admin", "password": "secret123", "host": "db.example.com"}
        secrets.upsert("database_credentials", credentials)

        # Retrieve secret
        db_creds = secrets.get("database_credentials")
        print(db_creds)

        {'username': 'admin', 'password': 'secret123', 'host': 'db.example.com'}

        # Delete secrets
        secrets.delete("database_credentials")
        ```
    """

    def __init__(self):
        """AWSSecretsManager Init Method"""
        self.log = logging.getLogger("workbench")

        # Initialize a Workbench Session (to assume the Workbench ExecutionRole)
        self.boto3_session = AWSSession().boto3_session

        # Create a Secrets Manager client
        self.secrets_client = self.boto3_session.client("secretsmanager")

    def list(self, prefix: str = None) -> list:
        """List all secrets in AWS Secrets Manager, optionally filtering by a prefix.

        Args:
            prefix (str, optional): A prefix to filter the secrets by. Defaults to None.

        Returns:
            list: A list of secret names.
        """
        try:
            # Set up parameters for the query
            params = {"MaxResults": 100}

            # Initialize the list to collect secret names
            all_secrets = []

            # Make the initial call to list secrets
            response = self.secrets_client.list_secrets(**params)

            # Filter and aggregate the names from the initial response
            for secret in response["SecretList"]:
                name = secret["Name"]
                if not prefix or name.startswith(prefix):
                    all_secrets.append(name)

            # Continue to paginate if there's a NextToken
            while "NextToken" in response:
                # Update the parameters with the NextToken for subsequent calls
                params["NextToken"] = response["NextToken"]
                response = self.secrets_client.list_secrets(**params)

                # Filter and aggregate the names from the subsequent responses
                for secret in response["SecretList"]:
                    name = secret["Name"]
                    if not prefix or name.startswith(prefix):
                        all_secrets.append(name)

        except Exception as e:
            self.log.error(f"Failed to list secrets: {e}")
            return []

        # Return the aggregated list of secret names
        return all_secrets

    def get(self, name: str, warn: bool = True) -> Union[str, list, dict, None]:
        """Retrieve a secret value from AWS Secrets Manager.

        Args:
            name (str): The name of the secret to retrieve.
            warn (bool): Whether to log a warning if the secret is not found.

        Returns:
            Union[str, list, dict, None]: The value of the secret or None if not found.
        """
        try:
            # Retrieve the secret from Secrets Manager
            response = self.secrets_client.get_secret_value(SecretId=name)

            # Get the secret value (could be in SecretString or SecretBinary)
            if "SecretString" in response:
                value = response["SecretString"]
            elif "SecretBinary" in response:
                value = response["SecretBinary"]
            else:
                self.log.warning(f"Secret '{name}' doesn't contain SecretString or SecretBinary")
                return None

            # Attempt to parse the value back to its original type
            try:
                parsed_value = json.loads(value)
                return parsed_value
            except (json.JSONDecodeError, TypeError):
                # If parsing fails, return the value as is (assumed to be a simple string)
                return value

        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                if warn:
                    self.log.warning(f"Secret '{name}' not found")
            else:
                self.log.error(f"Failed to get secret '{name}': {e}")
            return None

    def upsert(self, name: str, value: Union[str, list, dict]) -> bool:
        """Create or update a secret in AWS Secrets Manager.

        Args:
            name (str): The name of the secret to create or update.
            value (Union[str, list, dict]): The value to store.

        Returns:
            bool: True if successful, False otherwise.
        """
        try:
            # Convert complex types to JSON strings
            if not isinstance(value, str):
                value = json.dumps(value)

            # Check if the secret already exists
            try:
                self.secrets_client.describe_secret(SecretId=name)
                # Secret exists, update it
                self.secrets_client.update_secret(SecretId=name, SecretString=value)
                self.log.info(f"Updated secret '{name}'")
            except ClientError as e:
                if e.response["Error"]["Code"] == "ResourceNotFoundException":
                    # Secret doesn't exist, create it
                    self.secrets_client.create_secret(Name=name, SecretString=value)
                    self.log.info(f"Created secret '{name}'")
                else:
                    raise e

            return True

        except Exception as e:
            self.log.error(f"Failed to upsert secret '{name}': {e}")
            return False

    def delete(self, name: str, force=False) -> bool:
        """Delete a secret from AWS Secrets Manager.

        Args:
            name (str): The name of the secret to delete.
            force (bool): Whether to force delete the secret without recovery.

        Returns:
            bool: True if successful, False otherwise.
        """
        try:
            if force:
                # Force delete the secret without recovery
                self.secrets_client.delete_secret(SecretId=name, ForceDeleteWithoutRecovery=True)
                self.log.info(f"Force deleted secret '{name}' (not recoverable)")
                return True
            # Delete the secret with recovery (30 days)
            self.secrets_client.delete_secret(SecretId=name, ForceDeleteWithoutRecovery=False)
            self.log.info(f"Deleted secret '{name}' (recoverable for 30 days)")
            return True

        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                self.log.warning(f"Secret '{name}' not found for deletion")
                return True  # Consider it a success if it didn't exist anyway
            else:
                self.log.error(f"Failed to delete secret '{name}': {e}")
                return False


if __name__ == "__main__":
    """Exercise the AWS Secrets Manager class"""
    # Create an instance of the AWSSecretsManager
    secrets_manager = AWSSecretsManager()

    # List all secrets
    print("Listing all secrets:")
    print(secrets_manager.list())

    # Add a secret
    secrets_manager.upsert("test_api_key", "my-secret-key")
    print("Added secret 'test_api_key'")

    # Retrieve the secret
    api_key = secrets_manager.get("test_api_key")
    print(f"Retrieved secret 'test_api_key': {api_key}")

    # Add structured data
    credentials = {"username": "admin", "password": "secret123", "host": "db.example.com"}
    secrets_manager.upsert("test_database_credentials", credentials)
    print("Added secret 'test_database_credentials'")

    # Retrieve the structured data
    db_creds = secrets_manager.get("test_database_credentials")
    print(f"Retrieved secret 'test_database_credentials': {db_creds}")

    # Delete the secret
    secrets_manager.delete("test_database_credentials", force=True)
    print("Deleted secret 'test_database_credentials'")



"""An Endpoint Metrics Plots component"""

import math
from dash import dcc
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from datetime import datetime

# Workbench Imports
from workbench.web_interface.components.component_interface import ComponentInterface
from workbench.utils.pandas_utils import subplot_positions

# Get local timezone
local_tz = datetime.now().astimezone().tzinfo


class EndpointMetricPlots(ComponentInterface):
    """Endpoint Metrics Component"""

    def create_component(self, component_id: str) -> dcc.Graph:
        """Create a Endpoint Metrics Component without any data.
        Args:
            component_id (str): The ID of the web component
        Returns:
            dcc.Graph: The Endpoint Metrics Component
        """
        return dcc.Graph(
            id=component_id, figure=self.display_text("Waiting for Data..."), config={"displayModeBar": False}
        )

    def update_properties(self, endpoint_details: dict) -> go.Figure:
        """Create a Endpoint Metrics Figure for the numeric columns in the dataframe.
        Args:
            endpoint_details (dict): The model details dictionary
        Returns:
            plotly.graph_objs.Figure: A Figure object containing the Endpoint metrics.
        """

        # Grab the Endpoint metrics from the model details
        metrics_df = endpoint_details.get("endpoint_metrics")
        if metrics_df is None:
            return self.display_text("No Data")

        # Hack for super long metric names
        if "ServerlessConcurrentExecutionsUtilization" in metrics_df.columns:
            metrics_df.rename(
                columns={"ServerlessConcurrentExecutionsUtilization": "ConcurrentExecPercentage"}, inplace=True
            )

        # Let's convert all the timestamps to local timezone
        metrics_df["timestamps"] = metrics_df["timestamps"].dt.tz_convert(local_tz)

        # Move the timestamps to the index
        metrics_df.set_index("timestamps", inplace=True, drop=True)

        # Compute our subplot layout
        subplot_pos_lookup = subplot_positions(metrics_df)
        num_rows = math.ceil(len(metrics_df.columns) / 2)

        # Create the figure with subplots for each metric
        fig = make_subplots(rows=num_rows, cols=2, subplot_titles=metrics_df.columns, vertical_spacing=0.12)
        for metric in metrics_df.columns:
            row, col = subplot_pos_lookup[metric]
            fig.add_trace(go.Scatter(x=metrics_df.index, y=metrics_df[metric], fill="toself"), row=row, col=col)

        # Update the figure layout
        fig.update_xaxes(tickfont_size=10)
        fig.update_yaxes(rangemode="tozero", tickfont_size=10)
        fig.update_layout(
            showlegend=False,
            margin={"t": 30, "b": 20, "r": 10, "l": 30},
            height=500,
        )

        # Return the figure
        return fig


if __name__ == "__main__":
    # This class plots out Endpoint Metrics
    from workbench.core.artifacts.endpoint_core import EndpointCore

    # Grab the endpoint details
    end = EndpointCore("abalone-regression")
    end_details = end.details()

    # Instantiate the EndpointMetricPlots class
    endpoint_metric_plots = EndpointMetricPlots()

    # Generate the figure
    fig = endpoint_metric_plots.update_properties(end_details)

    # Apply dark theme
    fig.update_layout(template="plotly_dark")

    # Show the figure
    fig.show()



from dataclasses import dataclass, field
import logging
from typing import List

# Workbench Imports
from workbench.utils.chem_utils import svg_from_smiles


@dataclass
class Compound:
    """Compound: Store details about an individual compound."""

    id: str
    smiles: str = None
    tags: List[str] = field(default_factory=list)
    meta: dict = field(default_factory=dict)
    log: logging.Logger = field(default_factory=lambda: logging.getLogger("workbench"), init=False)

    def add_tag(self, tag: str) -> None:
        """Add a single tag to the tags list."""
        if tag not in self.tags:
            self.tags.append(tag)

    def remove_tag(self, tag: str) -> None:
        """Remove a single tag from the tags list."""
        if tag in self.tags:
            self.tags.remove(tag)

    def add_meta(self, key: str, value) -> None:
        """Add metadata to the Compound."""
        self.meta[key] = value

    def details(self) -> dict:
        """Compound Details

        Returns:
            dict: A dictionary of details about the Compound.
        """
        return {"project": "XYZ", "smiles": self.smiles, "tags": self.tags, "meta": self.meta}

    def image(self, width: int = 300, height: int = 200) -> str:
        """Generate an SVG image of the Compound

        Args:
            width (int, optional): The width of the image (default: 300)
            height (int, optional): The height of the image (default: 200)

        Returns:
            str: The SVG image of the Compound
        """
        from workbench.utils.theme_manager import ThemeManager  # Avoid circular import

        return svg_from_smiles(self.smiles, width, height, background=ThemeManager().background())

    def __str__(self) -> str:
        """User-friendly string representation."""
        str_output = (
            f"Compound({self.id})\n  SMILES: {self.smiles}\n  Tags: {', '.join(self.tags) if self.tags else 'None'}"
        )
        str_output += f"\n  Meta: {self.meta if self.meta else 'None'}"
        return str_output


if __name__ == "__main__":
    # Example usage
    compound = Compound("123")
    compound.smiles = "CCO"
    compound.tags = ["alcohol", "primary"]
    compound.add_meta("toxic_info", {"elements": ["C", "H", "O"], "groups": ["alcohol"]})

    # Print details of the Compound
    print(compound.details())

    # Debugging view (__repr__)
    print(repr(compound))

    # User-friendly view (__str__)
    print(compound)

    # Generate an SVG image of the Compound
    print(compound.image())



"""S3 Utilities for Workbench"""

import os
import io
import boto3
from urllib.parse import urlparse
import awswrangler as wr
from botocore.exceptions import ClientError
import hashlib
from typing import Optional
import logging

# Workbench imports
from workbench.utils.performance_utils import performance

log = logging.getLogger("workbench")


def upload_content_to_s3(content, path):
    """Write string content to S3 path"""
    bytes_io = io.BytesIO(content.encode("utf-8"))
    return wr.s3.upload(local_file=bytes_io, path=path)


def read_content_from_s3(path):
    """Read string content from S3 path"""
    buffer = io.BytesIO()
    wr.s3.download(path=path, local_file=buffer)
    buffer.seek(0)
    return buffer.read().decode("utf-8")


def get_s3_etag(s3_uri: str, session: boto3.session.Session) -> Optional[str]:
    """
    Retrieve the ETag of an S3 object.

    Args:
        s3_uri (str): The S3 URI of the object (e.g., 's3://bucket/key').
        session (boto3.session.Session): A boto3 session.

    Returns:
        Optional[str]: The ETag of the object if it exists, otherwise None.

    Note:
        In general AWS ETags aren't useful, they aren't content hashes, they just indicate
        'change' in the object, and for that you can just use the last modified date.
    """
    s3 = session.client("s3")

    try:
        # Parse bucket and key from the S3 URI
        bucket, key = s3_uri.replace("s3://", "").split("/", 1)
        response = s3.head_object(Bucket=bucket, Key=key)
        return response.get("ETag", "").strip('"')  # Remove quotes from ETag
    except s3.exceptions.ClientError:
        return None


def ensure_s3_bucket_and_prefix(s3_uri: str, session: boto3.session.Session):
    """
    Ensure the S3 bucket and prefix exist, creating them if necessary.

    Args:
        s3_uri (str): The S3 URI (e.g., 's3://bucket-name/prefix/').
        session (boto3.session.Session): The boto3 session.
    """
    s3 = session.client("s3")

    # Parse bucket and prefix from the S3 path
    bucket, *prefix_parts = s3_uri.replace("s3://", "").split("/", 1)
    prefix = prefix_parts[0] if prefix_parts else ""

    # Ensure bucket exists
    try:
        s3.head_bucket(Bucket=bucket)
    except ClientError as e:
        if e.response["Error"]["Code"] == "404":
            print(f"Creating bucket: {bucket}")
            s3.create_bucket(Bucket=bucket)
        else:
            raise e

    # Ensure prefix exists by creating a placeholder object
    if prefix:
        print(f"Ensuring prefix: {prefix}")
        s3.put_object(Bucket=bucket, Key=f"{prefix.rstrip('/')}/.placeholder", Body=b"")


def compute_s3_object_hash(s3_url: str, session: boto3.session.Session) -> str:
    """
    Compute the MD5 hash of an S3 object's content.

    Args:
        s3_url (str): The S3 URL (e.g., "s3://bucket-name/object-key").
        session (boto3.session.Session): Boto3 session.

    Returns:
        str: MD5 hash of the object's content.
    """
    log.important(f"Computing S3 Object Hash: {s3_url}")

    # Parse the S3 URL
    parsed_url = urlparse(s3_url)
    bucket_name = parsed_url.netloc
    object_key = parsed_url.path.lstrip("/")

    s3_client = session.client("s3")
    file_hash = hashlib.md5()
    response = s3_client.get_object(Bucket=bucket_name, Key=object_key)

    # Stream the object content to avoid memory overhead
    for chunk in response["Body"].iter_chunks(chunk_size=8192):
        file_hash.update(chunk)

    return file_hash.hexdigest()


@performance
def compute_parquet_hash(s3_url: str, session: boto3.session.Session) -> str:
    """
    Compute a composite content hash for a set of Parquet files in an S3 prefix.

    Args:
        s3_url (str): S3 URL for the FeatureGroup's offline storage (e.g., "s3://bucket-name/path/to/data/").
        session (boto3.session.Session): Boto3 session.

    Returns:
        str: Composite hash for a set of Parquet files.
    """
    log = logging.getLogger("workbench")
    s3_client = session.client("s3")

    # Parse bucket and prefix from the S3 URL
    parsed_url = urlparse(s3_url)
    bucket_name = parsed_url.netloc
    prefix = parsed_url.path.lstrip("/")

    # Ensure the prefix ends with a slash to match the exact directory
    if not prefix.endswith("/"):
        prefix += "/"

    # Initialize MD5 hash object for composite hash
    composite_hash = hashlib.md5()

    # Use paginator to iterate through objects in the S3 prefix
    paginator = s3_client.get_paginator("list_objects_v2")
    for page in paginator.paginate(Bucket=bucket_name, Prefix=prefix):
        for obj in page.get("Contents", []):
            if obj["Key"].startswith(prefix) and obj["Key"].endswith(".parquet"):
                # Compute the hash for the current Parquet file
                file_hash = compute_s3_object_hash(f"s3://{bucket_name}/{obj['Key']}", session)
                composite_hash.update(file_hash.encode("utf-8"))
                log.debug(f"Hash for {obj['Key']}: {file_hash}")

    return composite_hash.hexdigest()


def copy_s3_files_to_local(s3_path: str, local_path: str):
    """Copies all files from S3 to a local directory, maintaining the subdirectory structure.
    Args:
        s3_path (str): S3 Path to the set of files (e.g., s3://bucket-name/path/to/files).
        local_path (str): Local directory to copy the files to.
    """
    s3_client = boto3.client("s3")
    bucket, key = s3_path.replace("s3://", "").split("/", 1)

    paginator = s3_client.get_paginator("list_objects_v2")
    for page in paginator.paginate(Bucket=bucket, Prefix=key):
        for obj in page.get("Contents", []):
            # Correctly handle object keys to ensure paths are relative
            relative_key = obj["Key"][len(key) :].lstrip("/")  # Remove the S3 prefix and leading slashes
            local_file_path = os.path.join(local_path, relative_key)

            # Ensure the subdirectory structure exists locally
            os.makedirs(os.path.dirname(local_file_path), exist_ok=True)

            # Download the object to the local file path
            log.important(f"Downloading {bucket}/{obj['Key']} to {local_file_path}")
            s3_client.download_file(bucket, obj["Key"], local_file_path)


if __name__ == "__main__":
    """Exercise the S3 Utilities"""

    # Get our Account Clamp and S3 Bucket
    from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp
    from workbench.utils.config_manager import ConfigManager

    session = AWSAccountClamp().boto3_session
    workbench_bucket = ConfigManager().get_config("WORKBENCH_BUCKET")

    # Setup a temporary S3 prefix for the Athena output
    s3_scratch = f"s3://{workbench_bucket}/temp/s3_utils_test/"

    # Check if a bucket and prefix exist
    print(f"Ensuring bucket and prefix exist: {s3_scratch}")
    ensure_s3_bucket_and_prefix(s3_scratch, session)

    # Test the write and read functions
    test_string = "This is a test string."
    test_path = f"{s3_scratch}/test.txt"
    print(f"Writing to S3: {test_path}")
    upload_content_to_s3(test_string, test_path)
    print(f"Reading from S3: {test_path}")
    read_string = read_content_from_s3(test_path)
    print(f"Read string: {read_string}")
    assert read_string == test_string, "Read string does not match the original string."



"""RowTagger: Generalized Row Tagger (domain specific logic should be captured in a subclass"""

import pandas as pd

# Workbench Imports
from workbench.algorithms.dataframe import feature_spider


# Class: RowTagger
class RowTagger:
    """RowTagger: Encapsulate business logic and special cases for tagging rows in a dataframe
    - Zero Target Value
    - CoIncident (with reasonable difference in target value)
    - High Target Gradient (HTG) Neighborhood
    """

    def __init__(
        self,
        dataframe: pd.DataFrame,
        features: list,
        id_column: str,
        target_column: str,
        within_dist: float,
        min_target_diff: float,
        outlier_df: pd.DataFrame = None,
        categorical_target: bool = False,
    ):
        # Set up some parameters
        self.id_column = id_column
        self.target_column = target_column
        self.within_dist = within_dist
        self.min_target_diff = min_target_diff
        self.outlier_df = outlier_df

        # Do a validation check on the dataframe
        self.df = dataframe.copy()
        self.validate_input_data()

        # We need the feature spider for the more advanced tags
        self.f_spider = feature_spider.FeatureSpider(
            self.df,
            features,
            id_column=self.id_column,
            target_column=target_column,
            categorical_target=categorical_target,
        )

        # Add a 'tags' column (if it doesn't already exist)
        if "tags" not in self.df.columns:
            self.df["tags"] = [set() for _ in range(len(self.df.index))]

    def validate_input_data(self):
        # Make sure it's a dataframe
        if not isinstance(self.df, pd.DataFrame):
            print("Input is NOT a DataFrame!")
            return False

        # Make sure it has some rows and columns
        rows, columns = self.df.shape
        if rows == 0:
            print("Input DataFrame has 0 rows!")
            return False
        if columns == 0:
            print("Input DataFrame has 0 columns!")
            return False

        # Make sure it has an ID column (domain specific stuff in subclass)
        if self.id_column not in self.df.columns:
            print("Input DataFrame needs a ID column!")
            return False

        # AOK
        return True

    def tag_rows(self) -> pd.DataFrame:
        """Run all the current registered taggers"""
        taggers = [self.zero_target, self.coincident, self.high_gradients]
        for tagger in taggers:
            tagger()

        # Check for outliers
        if self.outlier_df is not None:
            # Merge the dataframes on the id_column, to align rows
            self.df = self.df.merge(self.outlier_df[[self.id_column]], on=self.id_column, how="left", indicator=True)

            # Use the merge indicator to add "outlier" tag to the relevant rows
            self.df["tags"] = self.df.apply(
                lambda row: row["tags"].union({"outlier"}) if row["_merge"] == "both" else row["tags"], axis=1
            )

            # Drop the merge indicator column
            self.df.drop(columns=["_merge"], inplace=True)

        # Convert the set to a string
        self.df["tags"] = self.df["tags"].apply(lambda x: ", ".join(x))
        return self.df

    def zero_target(self):
        """Find observations with a target value of 0"""
        zero_indexes = self.df[self.df[self.target_column] == 0].index
        for index in zero_indexes:
            self.df.at[index, "tags"].add("zero")

    def coincident(self):
        """Find observations with the SAME features that have different target values"""
        coincident_indexes = self.f_spider.coincident(self.min_target_diff, verbose=False)

        # We get back index offsets (not labels) so we need to use iloc
        for index in coincident_indexes:
            self.df["tags"].iloc[index].add("coincident")

    def high_gradients(self):
        """Find observations close in feature space with a high difference in target values
        High Target Gradient (HTG)"""
        htg_indexes = self.f_spider.high_gradients(self.within_dist, self.min_target_diff, verbose=False)

        # We get back index offsets (not labels) so we need to use iloc
        for index in htg_indexes:
            self.df["tags"].iloc[index].add("htg")


def simple_unit_test():
    """Unit Test for the RowTagger Class"""

    # Set some pandas options
    pd.set_option("display.max_columns", None)
    pd.set_option("display.width", 1000)

    # Make some fake data
    data = {
        "ID": [
            "id_0",
            "id_1",
            "id_2",
            "id_3",
            "id_4",
            "id_5",
            "id_6",
            "id_7",
            "id_8",
            "id_9",
        ],
        "feat1": [1.0, 1.0, 1.1, 3.0, 4.0, 1.0, 1.0, 1.1, 3.0, 4.0],
        "feat2": [1.0, 1.0, 1.1, 3.0, 4.0, 1.0, 1.0, 1.1, 3.0, 4.0],
        "feat3": [0.1, 0.1, 0.2, 1.6, 2.5, 0.1, 0.1, 0.2, 1.6, 2.5],
        "price": [31, 60, 62, 40, 20, 31, 61, 60, 40, 0],
    }
    data_df = pd.DataFrame(data)

    # Create the class and run the taggers
    row_tagger = RowTagger(
        data_df,
        features=["feat1", "feat2", "feat3"],
        id_column="ID",
        target_column="price",
        within_dist=2.0,
        min_target_diff=1.0,
    )
    data_df = row_tagger.tag_rows()
    print(data_df)


def unit_test():
    """Unit Test for the FeatureResolution Class"""
    from workbench.api.feature_set import FeatureSet
    from workbench.api.model import Model

    # Grab a test dataframe
    fs = FeatureSet("aqsol_mol_descriptors")
    test_df = fs.pull_dataframe()

    # Get the target and feature columns
    m = Model("aqsol-mol-regression")
    target_column = m.target()
    feature_columns = m.features()

    row_tagger = RowTagger(
        test_df,
        features=feature_columns,
        id_column="id",
        target_column=target_column,
        within_dist=0.25,
        min_target_diff=1.0,
        outlier_df=fs.outliers(),
    )
    data_df = row_tagger.tag_rows()
    print(data_df["tags"].value_counts())


if __name__ == "__main__":
    simple_unit_test()
    unit_test()



"""This Script creates a simple AWS ML Pipeline with Workbench

DataSources:
    - abalone_data
FeatureSets:
    - abalone_features
Models:
    - abalone-regression
Endpoints:
    - abalone-regression-end
"""

import sys
from pathlib import Path
from workbench.core.artifacts.model_core import ModelType
from workbench.core.transforms.data_loaders.light.csv_to_data_source import CSVToDataSource
from workbench.core.transforms.data_to_features.light.data_to_features_light import DataToFeaturesLight
from workbench.core.transforms.features_to_model.features_to_model import FeaturesToModel
from workbench.core.transforms.model_to_endpoint.model_to_endpoint import ModelToEndpoint


if __name__ == "__main__":
    # Get the path to the dataset in the repository data directory
    abalone_data_path = Path(sys.modules["workbench"].__file__).parent.parent.parent / "data" / "abalone.csv"

    # Create the abalone_data DataSource
    my_loader = CSVToDataSource(abalone_data_path, "abalone_data")
    my_loader.set_output_tags(["abalone", "public"])
    my_loader.transform()

    # Create the abalone_features FeatureSet
    data_to_features = DataToFeaturesLight("abalone_data", "abalone_features")
    data_to_features.set_output_tags(["abalone", "public"])
    data_to_features.transform()

    # Create the abalone_regression Model
    features_to_model = FeaturesToModel("abalone_features", "abalone-regression", ModelType.REGRESSOR)
    features_to_model.set_output_tags(["abalone", "regression"])
    features_to_model.transform(target_column="class_number_of_rings", description="Abalone Regression Model")

    # Create the abalone_regression Endpoint
    model_to_endpoint = ModelToEndpoint("abalone-regression", "abalone-regression")
    model_to_endpoint.set_output_tags(["abalone", "regression"])
    model_to_endpoint.transform()
